from collections import defaultdict
from typing import Generator, Optional, TypeVar, Union, TYPE_CHECKING

from dliswriter.logical_record.core.eflr import EFLRSet

if TYPE_CHECKING:
    from dliswriter.logical_record.core.eflr.eflr_item import EFLRItem


AnyEFLRSet = TypeVar("AnyEFLRSet", bound="EFLRSet")
AnyEFLRItem = TypeVar("AnyEFLRItem", bound="EFLRItem")


class EFLRSetsDict(defaultdict):
    """Keep track of the defined EFLRSet objects.

    This is a nested dictionary.
    At the top level, it is a defaultdict whose keys are EFLRSet classes. The values are simple dicts.
    At the second level, the keys are EFLRSet instances names and the values are the corresponding instances.
    """

    def __init__(self) -> None:
        """Initialise EFLRSetsDict. Set it up to return an empty dictionary for a new (missing) key."""

        super().__init__(lambda: {})

    def add_set(self, eflr_set: EFLRSet) -> None:
        """Register a new EFLRSet instance in the structure."""

        if eflr_set.set_name in self[eflr_set.__class__]:
            raise RuntimeError(f"{eflr_set.__class__.__name__} with set name '{eflr_set.set_name}' "
                               f"already added to the file")

        self[eflr_set.__class__][eflr_set.set_name] = eflr_set

    def try_add_set(self, eflr_set: EFLRSet) -> bool:
        """Try to register a new EFLRSet instance in the structure. Return True on success, False otherwise."""

        if eflr_set.set_name in self[eflr_set.__class__]:
            return False
        else:
            self[eflr_set.__class__][eflr_set.set_name] = eflr_set
            return True

    def get_or_make_set(self, eflr_set_type: type[AnyEFLRSet], set_name: Optional[str] = None) -> AnyEFLRSet:
        """Given an EFLRSet subclass and name, either retrieve a relevant EFLRSet from the structure or create it.

        Args:
            eflr_set_type   :   Subclass of EFLRSet - type of the EFLRSet to be obtained.
            set_name        :   Name of the EFLRSet instance.

        Returns:
            An EFLRSet instance of given subtype and name, registered in the structure.
        """

        # dict mapping set names on EFLRSet (subclass) instances
        eflr_set_dict: dict[Union[str, None], AnyEFLRSet] = self[eflr_set_type]

        # instance of the EFLRSet with the given set name - if exists
        eflr_set_instance = eflr_set_dict.get(set_name, None)

        # (if not exists - create it)
        if eflr_set_instance is None:
            eflr_set_instance = eflr_set_type(set_name=set_name)
            eflr_set_dict[set_name] = eflr_set_instance

        return eflr_set_instance

    def get_all_items_for_set_type(self, eflr_set_type: type[EFLRSet]) -> Generator[AnyEFLRItem, None, None]:
        """Retrieve all EFLRItem instances registered for all instances of given EFLRSet subclass."""

        for value in self[eflr_set_type].values():
            yield from value.get_all_eflr_items()



from .attribute import Attribute
from .subtypes import (NumericAttribute, DimensionAttribute, DTimeAttribute, EFLRAttribute, StatusAttribute,
                       TextAttribute, IdentAttribute, EFLROrTextAttribute, PropertiesAttribute)



from dlisio import dlis    # type: ignore  # untyped library


def test_long_name_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Test attributes of the LongName object in the DLIS file."""

    w = short_dlis.longnames[0]
    t = 'SOME ASCII TEXT'

    assert w.modifier == [t]
    assert w.quantity == t
    assert w.quantity_mod == [t]
    assert w.altered_form == t
    assert w.entity == t
    assert w.entity_mod == [t]
    assert w.entity_nr == t
    assert w.entity_part == t
    assert w.entity_part_nr == t
    assert w.generic_source == t
    assert w.source_part == [t]
    assert w.source_part_nr == [t]
    assert w.conditions == [t]
    assert w.standard_symbol == t
    assert w.private_symbol == t
    assert w.origin == 42



import re
from typing import Union

from dliswriter.configuration import global_config


HC_STRING_PATTERN = re.compile(r"[A-Z0-9_-]+")


def validate_string(s: str) -> str:
    """Check that a strings fulfills the conditions to be deemed safe for all known DLIS viewers.

    For more details, see https://well-id-widcdliswriter.readthedocs-hosted.com/userguide/compatibilityissues.html

    Returns the original string.
    """

    if not isinstance(s, str):
        raise TypeError(f"Expected a str, got {type(s)}: {s}")

    if not global_config.high_compat_mode:
        return s

    if HC_STRING_PATTERN.fullmatch(s) is None:
        raise ValueError("In high-compatibility mode, strings can contain only uppercase characters, digits, "
                         f"dashes, and underscores; got {repr(s)}")

    return s


def convert_numeric(value: str) -> Union[int, float, str]:
    """Convert a string to an integer or float."""

    parser = float if '.' in value else int

    try:
        parsed_value = parser(value)
    except ValueError:
        raise ValueError(f"Value '{value}' could not be converted to a numeric type")
    return parsed_value


def convert_maybe_numeric(val: Union[str, int, float]) -> Union[str, int, float]:
    """Try converting a value to a number. If that fails, return the value unchanged."""

    if isinstance(val, (int, float)):
        return val

    if not isinstance(val, str):
        raise TypeError(f"Expected an int, float, or str; got {type(val)}: {val}")
    try:
        return convert_numeric(val)
    except ValueError:
        pass
    return val



import pytest
from dlisio import dlis    # type: ignore  # untyped library


@pytest.mark.parametrize(("idx", "name", "consumer_name", "description"), (
        (0, "no_format_1", "SOME TEXT NOT FORMATTED", "TESTING-NO-FORMAT"),
        (1, "no_fmt2", "xyz", "TESTING NO FORMAT 2")
))
def test_no_format_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, consumer_name: str,
                          description: str) -> None:
    """Check attributes of NoFormat objects in the DLIS file."""

    w = short_dlis.noformats[idx]

    assert w.name == name
    assert w.consumer_name == consumer_name
    assert w.description == description
    assert w.origin == 42



from typing import Any

from dliswriter.utils.internal.converters import get_ascii_bytes
from dliswriter.logical_record.core.logical_record import LogicalRecordBytes
from dliswriter.utils.internal.value_checkers import validate_string


class StorageUnitLabel:
    """Model a Storage Unit Label - the first part of a DLIS file.

    Storage Unit Label is always 80 bytes long.

    Format:
        First 4 bytes:  Storage Unit Sequence Number
        Next 5 bytes:   DLIS version
        Next 6 bytes:   Storage Unit Structure
        Next 5 bytes:   Maximum Record Length
        Next 60 bytes:  Storage Set Identifier

    Although the StorageUnitLabel is not technically a LogicalRecord subclass, its interface has been adjusted so that
    it mocks the latter - for easier integration in the file writer.

    """

    storage_unit_structure = 'RECORD'   #: storage unit structure (this is the only allowed value)
    dlis_version = 'V1.00'              #: version of DLIS this implementation is for
    max_record_length_limit = 16384     #: maximal allowed length of a visible record

    def __init__(self, set_identifier: str, sequence_number: int = 1, max_record_length: int = 8192):
        """Initialise StorageUnitLabel.

        Args:
            sequence_number     :   Indicates the order in which the current Storage Unit appears in a Storage Set.
            set_identifier      :   ID of the storage set (e.g. "Default Storage Set").
            max_record_length   :   Maximum length of each visible record;
                                    see  # http://w3.energistics.org/rp66/v1/rp66v1_sec2.html#2_3_6_5
        """

        super().__init__()

        self.sequence_number = sequence_number
        self.set_identifier = validate_string(set_identifier)
        self.max_record_length = max_record_length

        if max_record_length > self.max_record_length_limit:
            raise ValueError(f"Max record length cannot be larger than {self.max_record_length_limit}")

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}(sequence_number={self.sequence_number}, "
                f"set_identifier={self.set_identifier}, max_record_length={self.max_record_length})")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if not self.sequence_number == other.sequence_number:
            return False

        if not self.set_identifier == other.set_identifier:
            return False

        if not self.max_record_length == other.max_record_length:
            return False

        return True

    def represent_as_bytes(self) -> LogicalRecordBytes:
        """Create bytes describing the storage unit label.

        Returns:
            Bytes of complete Storage Unit Label, wrapped in a LogicalRecordBytes object for consistency with
            Logical Records - other objects handled in the same loops as the SUL.
        """

        # Storage Unit Sequence Number
        _susn_as_bytes = get_ascii_bytes(str(self.sequence_number), 4)

        # DLIS Version
        _dlisv_as_bytes = get_ascii_bytes(self.dlis_version, 5, justify_left=True)

        # Storage Unit Structure
        _sus_as_bytes = get_ascii_bytes(self.storage_unit_structure, 6)

        # Maximum Record Length
        _mrl_as_bytes = get_ascii_bytes(str(self.max_record_length), 5)

        # Storage Set Identifier
        _ssi_as_bytes = get_ascii_bytes(self.set_identifier, 60, justify_left=True)

        bts = _susn_as_bytes + _dlisv_as_bytes + _sus_as_bytes + _mrl_as_bytes + _ssi_as_bytes

        return LogicalRecordBytes(bts, lr_type_struct=b'')



import pytest
from dlisio import dlis    # type: ignore  # untyped library


@pytest.mark.parametrize(("idx", "name", "v_zero", "m_decl", "c1_name", "c1_value", "c2_name", "c2_value"), (
        (0, "AQLN WELL-REF", "AQLN vertical_zero", 999.51, "Latitude", 40.395240, "Longitude", 27.792470),
        (1, "WRP-X", "vz20", 112.3, "X", 20, "Y", -0.3)
))
def test_well_reference_point_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, v_zero: str,
                                     m_decl: float, c1_name: str, c1_value: float, c2_name: str,
                                     c2_value: float) -> None:
    """Check attributes of well reference point in the DLIS file."""

    w = short_dlis.wellrefs[idx]

    assert w.name == name
    assert w.vertical_zero == v_zero
    assert w.magnetic_declination == m_decl
    assert w.coordinates[c1_name] == c1_value
    assert w.coordinates[c2_name] == c2_value
    assert w.origin == 42



from dliswriter.utils.internal.internal_enums import IFLRType
from dliswriter.logical_record.core.logical_record import LogicalRecord


class IFLR(LogicalRecord):
    """Model an Indirectly Formatted Logical Record.

    This type of logical record is mainly mean for numerical data.

    This is an abstract base class; see its subclasses (FrameData and NoFormatFrameData) for final implementations.
    """

    logical_record_type: IFLRType   #: int-enum denoting type of the EFLR
    is_eflr = False                 #: indication that this is an indirectly formatted LR

    def __init__(self) -> None:
        """Initialise an IFLR."""

        super().__init__()



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRItem, EFLRSet, DimensionedItem
from dliswriter.logical_record.eflr_types.axis import AxisSet
from dliswriter.logical_record.eflr_types.zone import ZoneSet
from dliswriter.logical_record.eflr_types.long_name import LongNameSet
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import (EFLRAttribute, NumericAttribute, DimensionAttribute,
                                                      EFLROrTextAttribute, PropertiesAttribute)


logger = logging.getLogger(__name__)


class ComputationItem(EFLRItem, DimensionedItem):
    """Model an object being part of Computation EFLR."""

    parent: "ComputationSet"

    def __init__(self, name: str, parent: "ComputationSet", **kwargs: Any) -> None:
        """Initialise ComputationItem.

        Args:
            name            :   Name of the ComputationItem.
            parent          :   Parent ComputationSet of this ComputationItem.
            **kwargs        :   Values of to be set as characteristics of the ComputationItem Attributes.
        """

        self.long_name = EFLROrTextAttribute('long_name', object_class=LongNameSet)
        self.properties = PropertiesAttribute('properties')
        self.dimension = DimensionAttribute('dimension')
        self.axis = EFLRAttribute('axis', object_class=AxisSet, multivalued=True)
        self.zones = EFLRAttribute('zones', object_class=ZoneSet, multivalued=True)
        self.values = NumericAttribute('values', multivalued=True, multidimensional=True)
        self.source = EFLRAttribute('source')

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Set up default values of ComputationItem parameters if not explicitly set previously."""

        if self.values.value is not None and self.zones.value is not None:
            if (nv := len(self.values.value)) != (nz := self.zones.count):
                raise RuntimeError("A Computation must have the same number of values and zones if both are "
                                   f"defined; got {nv} values and {nz} zones in {self}")

        self._check_axis_vs_dimension()
        self._check_or_set_value_dimensionality(self.values.value)

        if self.values.value and not self.dimension.value:
            logger.debug(f"Setting dimension of '{self}' to [1]")
            self.dimension.value = [1]


class ComputationSet(EFLRSet):
    """Model Computation EFLR."""

    set_type = 'COMPUTATION'
    logical_record_type = EFLRType.STATIC
    item_type = ComputationItem


ComputationItem.parent_eflr_class = ComputationSet



from dliswriter import DLISFile, StorageUnitLabel
from dliswriter.logical_record.eflr_types import FileHeaderSet, FileHeaderItem


def make_file_header() -> FileHeaderItem:
    return FileHeaderItem("DEFAULT FHLR", sequence_number=1, parent=FileHeaderSet())


def make_sul() -> StorageUnitLabel:
    return StorageUnitLabel("DEFAULT STORAGE SET", sequence_number=1)


def make_df() -> DLISFile:
    df = DLISFile(storage_unit_label=make_sul())

    lf = df.add_logical_file(file_header=make_file_header())

    lf.add_origin(
        "DEFINING ORIGIN",
        creation_time="2050/03/02 15:30:00",
        file_set_number=1,
        origin_reference=1,
    )
    return df



import pytest
from typing import Optional

from dliswriter.logical_record.eflr_types.path import PathItem, PathSet
from dliswriter.logical_record.eflr_types.frame import FrameItem
from dliswriter.logical_record.eflr_types.well_reference_point import WellReferencePointItem


@pytest.mark.parametrize(("name", "frm", "value", "wrp", "vertical_depth", "time"), (
        ("Path 1", "frame", ["channel1", "channel3"], "well_reference_point", 10., 12.5),
        ("PATH2", None, ["channel2"], None, -123.9, 277.),
))
def test_path_params(name: str, frm: Optional[str], value: list[str], wrp: Optional[str],
                     vertical_depth: float, time: float, request: pytest.FixtureRequest,
                     frame: FrameItem, well_reference_point: WellReferencePointItem) -> None:
    """Test creating PathItem."""

    path = PathItem(
        name,
        frame_type=request.getfixturevalue(frm) if frm else None,
        value=[request.getfixturevalue(v) for v in value],
        well_reference_point=request.getfixturevalue(wrp) if wrp else None,
        vertical_depth=vertical_depth,
        time=time,
        parent=PathSet()
    )

    assert path.name == name

    for i, n in enumerate(value):
        assert path.value.value[i] is request.getfixturevalue(value[i])

    if frm is not None:
        assert path.frame_type.value is frame

    if wrp is not None:
        assert path.well_reference_point.value is well_reference_point

    assert isinstance(path.parent, PathSet)
    assert path.parent.set_name is None



import pytest
from datetime import datetime, timezone
from dlisio import dlis    # type: ignore  # untyped library
from typing import Any


def test_zones(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of zones in the file matches the expected one."""

    zones = short_dlis.zones
    assert len(zones) == 5


@pytest.mark.parametrize(("name", "description", "maximum", "minimum", "value_type"), (
        ("Zone-1", "BOREHOLE-DEPTH-ZONE", 1300, 100, float),
        ("Zone-2", "VERTICAL-DEPTH-ZONE", 2300.45, 200, float),
        ("Zone-3", "ZONE-TIME", datetime(2050, 7, 13, 11, 30).astimezone(timezone.utc),
         datetime(2050, 7, 12, 9).astimezone(timezone.utc), datetime),
        ("Zone-4", "ZONE-TIME-2", 90, 10, float),
        ("Zone-X", "Zone not added to any parameter", 10, 1, float)
))
def test_zone_params(short_dlis: dlis.file.LogicalFile, name: str, description: str, maximum: Any,
                     minimum: Any, value_type: type) -> None:
    """Check attributes of zones in the new DLIS file."""

    zones = [zone for zone in short_dlis.zones if zone.name == name]
    assert len(zones) == 1
    z = zones[0]

    assert z.description == description

    assert isinstance(z.maximum, value_type)
    assert isinstance(z.minimum, value_type)

    z_maximum = z.maximum
    z_minimum = z.minimum

    if value_type is datetime:
        # dlisio doesn't add time zone info to the parsed datetime objects; mark them as UTC using .replace()
        z_maximum = z_maximum.replace(tzinfo=timezone.utc)  # type: ignore
        z_minimum = z_minimum.replace(tzinfo=timezone.utc)  # type: ignore

    assert z_maximum == maximum
    assert z_minimum == minimum

    assert z.origin == 42


def test_zone_not_in_param(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that a zone which has not been added to any parameter or other object is still in the file."""

    name = 'Zone-X'
    z = [z for z in short_dlis.zones if z.name == name]
    assert len(z) == 1
    for p in short_dlis.parameters:
        z = [z for z in p.zones if z.name == name]
        assert not z



import pytest
from dlisio import dlis    # type: ignore  # untyped library

from tests.common import check_list_of_objects


def test_tool(short_dlis: dlis.file.LogicalFile) -> None:
    tools = short_dlis.tools
    assert len(tools) == 2


@pytest.mark.parametrize(("idx", "name", "description", "status", "param_names", "channel_names"), (
        (0, "TOOL-1", "SOME TOOL", 1, ["Param-1", "Param-3"], ["posix time", "amplitude"]),
        (1, "Tool-X", "desc", 0, ["Param-2"], ["radius_pooh"])
))
def test_tool_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, description: str, status: int,
                     param_names: list[str], channel_names: list[str]) -> None:
    tool = short_dlis.tools[idx]
    assert tool.name == name
    assert tool.description == description
    assert tool.status == status
    assert tool.origin == 42

    check_list_of_objects(tool.parameters, param_names)
    check_list_of_objects(tool.channels, channel_names)



import numpy as np
import pytest
from typing import Any
from datetime import datetime, timedelta
from enum import Enum
from struct import Struct

from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter.utils.internal.converters import ReprCodeConverter
from dliswriter.utils.internal.types import numpy_dtype_type


@pytest.mark.parametrize(('dt', 'rc'), (
        (np.float64, RepresentationCode.FDOUBL),
        (np.float32, RepresentationCode.FSINGL),
        (np.int32, RepresentationCode.SLONG),
        (np.int16, RepresentationCode.SNORM),
        (np.int8, RepresentationCode.SSHORT),
        (np.uint32, RepresentationCode.ULONG),
        (np.uint16, RepresentationCode.UNORM),
        (np.uint8, RepresentationCode.USHORT),
))
def test_determine_repr_code_from_numpy_dtype(dt: numpy_dtype_type, rc: RepresentationCode) -> None:
    assert ReprCodeConverter.determine_repr_code_from_numpy_dtype(dt) is rc
    assert ReprCodeConverter.determine_repr_code_from_numpy_dtype(np.dtype(dt)) is rc


@pytest.mark.parametrize(('t', 'rc'), (
        (int, RepresentationCode.SLONG),
        (float, RepresentationCode.FDOUBL),
        (str, RepresentationCode.ASCII),
        (datetime, RepresentationCode.DTIME)
))
def test_determine_repr_code_from_generic_type(t: type, rc: RepresentationCode) -> None:
    assert ReprCodeConverter.determine_repr_code_from_generic_type(t) is rc


@pytest.mark.parametrize('t', (object, Enum, np.ndarray, Struct, dict, set, list, tuple))
def test_cannot_determine_repr_code_from_type(t: type) -> None:
    with pytest.raises(ReprCodeConverter.ReprCodeError, match="Cannot determine representation code for type.*"):
        ReprCodeConverter.determine_repr_code_from_generic_type(t)


@pytest.mark.parametrize('t', (1, 123.1, [1, 2, 4], 'abc'))
def test_determine_repr_code_from_generic_type_not_a_type(t: Any) -> None:
    with pytest.raises(TypeError, match=".* is not a type"):
        ReprCodeConverter.determine_repr_code_from_generic_type(t)


@pytest.mark.parametrize(('v', 'rc'), (
        (datetime.now(), RepresentationCode.DTIME),
        (2, RepresentationCode.SLONG),
        (-10208329, RepresentationCode.SLONG),
        (-92003198.2, RepresentationCode.FDOUBL),
        (3.123121231, RepresentationCode.FDOUBL),
        ('abc', RepresentationCode.ASCII),
        (np.arange(start=0, stop=10, dtype=np.int32), RepresentationCode.SLONG),
        (np.random.rand(12, 13), RepresentationCode.FDOUBL)
))
def test_determine_repr_code_from_value_single(v: Any, rc: RepresentationCode) -> None:
    assert ReprCodeConverter.determine_repr_code_from_value(v) is rc


@pytest.mark.parametrize(('v', 'rc'), (
        ([1, 2, 3], RepresentationCode.SLONG),
        ((1, 2, 3.5), RepresentationCode.FDOUBL),
        ('abcd', RepresentationCode.ASCII),
        (['a', 'b', 'c'], RepresentationCode.ASCII),
        ([datetime.now(), datetime.now() + timedelta(seconds=5)], RepresentationCode.DTIME)
))
def test_determine_repr_code_from_value_multiple(v: Any, rc: RepresentationCode) -> None:
    assert ReprCodeConverter.determine_repr_code_from_value(v) is rc


@pytest.mark.parametrize('v', (
    [0, 1, datetime.now()],
    [1, '1'],
    ('0.1', 0.2, '0.3')
))
def test_cannot_determine_common_repr_code(v: Any) -> None:
    with pytest.raises(ReprCodeConverter.ReprCodeError, match="Cannot determine a common representation code"):
        ReprCodeConverter.determine_repr_code_from_value(v)


@pytest.mark.parametrize(('rc', 'dt'), (
        (RepresentationCode.FDOUBL, np.float64),
        (RepresentationCode.FSINGL, np.float32),
        (RepresentationCode.SLONG, np.int32),
        (RepresentationCode.SNORM, np.int16),
        (RepresentationCode.SSHORT, np.int8),
        (RepresentationCode.ULONG, np.uint32),
        (RepresentationCode.UNORM, np.uint16),
        (RepresentationCode.USHORT, np.uint8)
))
def test_determine_numpy_dtype_from_repr_code(rc: RepresentationCode, dt: type[np.generic]) -> None:
    assert ReprCodeConverter.determine_numpy_dtype_from_repr_code(rc) == dt



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'dliswriter'
copyright = '2023, Well ID'
author = 'Dominika Dlugosz, Magne Lauritzen, Kamil Grunwald, Omer Faruk Sari'
release = '1.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions: list[str] = [
    'sphinxcontrib.mermaid'
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'classic'
html_static_path = ['_static']
html_logo = "../src/dliswriter/logo.png"

# Add an "Edit on GitHub" link to the top of all pages
html_context = {
    "display_github": True,
    "github_user": "well-id",
    "github_repo": "dliswriter",
    "github_version": "main",
    "conf_py_path": "/docs/",
}



import logging
import numpy as np
from typing import Union, Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode as RepC
from dliswriter.utils.enums import FrameIndexType
from dliswriter.logical_record.eflr_types.channel import ChannelSet, ChannelItem
from dliswriter.logical_record.core.attribute import (Attribute, EFLRAttribute, NumericAttribute, TextAttribute,
                                                      IdentAttribute)
from dliswriter.utils.source_data_wrappers import SourceDataWrapper
from dliswriter.configuration import global_config


logger = logging.getLogger(__name__)


class FrameItem(EFLRItem):
    """Model an object being part of Frame EFLR."""

    parent: "FrameSet"

    def __init__(self, name: str, parent: "FrameSet", **kwargs: Any) -> None:
        """Initialise FrameItem.

        Args:
            name        :   Name of the FrameItem.
            parent      :   Parent FrameSet of this FrameItem.
            **kwargs    :   Values of to be set as characteristics of the FrameItem Attributes.
        """

        self.description = TextAttribute('description')
        self.channels = EFLRAttribute('channels', object_class=ChannelSet, multivalued=True)
        self.index_type = IdentAttribute(
            'index_type',
            converter=FrameIndexType.make_converter("index types", soft=True)
        )
        self.direction = IdentAttribute('direction')
        self.spacing = NumericAttribute('spacing')
        self.encrypted = NumericAttribute(
            'encrypted', converter=self.convert_encrypted, representation_code=RepC.USHORT)
        self.index_min = NumericAttribute('index_min')
        self.index_max = NumericAttribute('index_max')

        super().__init__(name, parent=parent, **kwargs)

    @staticmethod
    def convert_encrypted(value: Union[str, int, float, bool]) -> int:
        """Convert a provided 'encrypted' attribute value to an integer flag (0 or 1)."""

        if isinstance(value, bool):
            return int(value)
        if isinstance(value, str):
            if value.lower() in ('1', 'true', 't', 'yes', 'y'):
                return 1
            elif value.lower() in ('0', 'false', 'f', 'no', 'n'):
                return 0
            else:
                raise ValueError(f"Couldn't evaluate the boolean meaning of '{value}'")
        if isinstance(value, (int, float)):
            if value != 1 and value != 0:
                raise ValueError(f"Expected a 0 or a 1; got {value}")
            return int(value)
        else:
            raise TypeError(f"Cannot convert {type(value)} object ({value}) to integer")

    def setup_from_data(self, data: SourceDataWrapper) -> None:
        """Set up attributes of the frame and its channels based on the source data."""

        if not self.channels.value:
            raise RuntimeError(f"No channels defined for {self}")

        for channel in self.channels.value:
            channel.set_dimension_and_repr_code_from_data(data)

        self._setup_frame_params_from_data(data)

    def _setup_frame_params_from_data(self, data: SourceDataWrapper) -> None:
        """Set up the index characteristics of the frame based on the source data.

        The index characteristics include: min and max value, spacing, and direction (increasing/decreasing).

        This method assumes that the first channel added to the frame is the index channel.
        This assumption is frequently made in DLIS readers.
        """

        def assign_if_none(attr: Attribute, value: Any, key: str = 'value') -> None:
            """Check if an attribute part has already been assigned. If not, assign it to the provided value.

            Args:
                attr    :   Attribute instance whose part should be assigned a value.
                value   :   Value to be assigned to the attribute part, if no value has been assigned so far.
                key     :   Name of the part of attribute which should be assigned.
            """

            if getattr(attr, key) is None and value is not None:
                logger.debug(f"Setting {attr.label}.{key} of {self} to {value}")
                setattr(attr, key, value)

        index_channel: ChannelItem = self.channels.value[0]
        index_data = data[index_channel.name][:]

        if self.index_type.value is None:
            # according to RP66, if index_type is None:
            #   - spacing and direction are meaningless - b ut some viewer software need a spacing defined
            #   - there is no index channel, so index_min and index_max should be frame-data numbers (row numbers)
            logger.info(f"No index channel defined for {self}; it will be indexed by the row number")
            assign_if_none(self.spacing, 1)
            assign_if_none(self.index_min, 1)
            assign_if_none(self.index_max, index_data.shape[0])

        else:
            assign_if_none(self.index_min, index_data.min())
            assign_if_none(self.index_max, index_data.max())
            for at in (self.index_min, self.index_max, self.spacing):
                assign_if_none(at, key='units', value=index_channel.units.value)

            if index_data.ndim != 1:
                raise RuntimeError(f"Index channel's data must be 1-dimensional; got {index_data.ndim} dimensions "
                                   f"for {index_channel} of {self}")
            spacing, direction = self._compute_spacing_and_direction(index_data)

            if spacing is None:
                # spacing cannot be used because it is not uniform enough; using only direction - if available
                m = (f"Spacing of the index channel of {self} is not uniform; this can cause issues in some viewer "
                     f"software. Consider implicit indexing by row number number instead "
                     f"by removing frame index_type specification.")
                if global_config.high_compat_mode:
                    raise RuntimeError(m)
                logger.warning(m)
                if direction is not None:
                    assign_if_none(self.direction, 'INCREASING' if direction > 0 else 'DECREASING')
            else:
                assign_if_none(self.spacing, spacing)
                # no need to define direction if spacing is defined

    @staticmethod
    def _compute_spacing_and_direction(index_data: np.ndarray) -> tuple[Union[int, float, None], Union[bool, None]]:
        """Compute spacing and direction of the data.

        Note:
            If spacing is not uniform enough, it is assigned to None.
            If direction cannot be determined, it is assigned to None.
        """

        diff = np.diff(index_data)
        diff_unique = np.unique(diff)

        if (diff_unique == 0).all():
            direction = None  # all zeros ->not determined
        elif (diff_unique >= 0).all():
            direction = True  # all non-negative, at least one positive -> increasing
        elif (diff_unique <= 0).all():
            direction = False  # all non-positive, at least one negative -> decreasing
        else:
            direction = None  # some increasing, some decreasing -> not determined

        if len(diff_unique) == 1:
            # if spacing between each sample is the same, this is it
            return diff_unique[0], direction

        # if not, check if these are minor deviations (can be attributed to numerical accuracy) or not
        median_diff = np.median(diff).item()  # mypy complains that median_diff is a numpy array...
        if median_diff == 0:
            return None, direction  # need the median for denominator later; if it's 0, cannot determine uniformity

        deviations = (1 - diff_unique / median_diff) ** 2
        if (deviations < 0.001).all():  # if differences are small enough, median is representative for spacing
            return median_diff, direction

        # differences too big for spacing to be assumed uniform; set spacing to None
        return None, direction

    @property
    def channel_name_mapping(self) -> dict:
        """Mapping of names of channels of the frame on the names of the associated datasets."""

        return {ch.name: ch.dataset_name for ch in self.channels.value}

    @property
    def known_channel_dtypes_mapping(self) -> dict:
        """Mapping of names of channels of the frame on the data types, if explicitly defined."""

        return {ch.name: ch.cast_dtype for ch in self.channels.value if ch.cast_dtype is not None}


class FrameSet(EFLRSet):
    """Model Frame EFLR."""

    set_type = 'FRAME'
    logical_record_type = EFLRType.FRAME
    item_type = FrameItem


FrameItem.parent_eflr_class = FrameSet



from dliswriter.file.file import DLISFile

from tests.dlis_files_for_testing.common import make_sul, make_file_header


def test_adding_origin_after_other_objects() -> None:
    df = DLISFile(storage_unit_label=make_sul())
    lf = df.add_logical_file(file_header=make_file_header())

    ch_depth = lf.add_channel(name="depth")
    ch_rpm = lf.add_channel(name="surface rpm")

    assert lf.default_origin_reference is None
    origin = lf.add_origin("DEFINING ORIGIN", file_set_number=38)
    assert lf.default_origin_reference == 0

    ch_time = lf.add_channel("time")
    frame = lf.add_frame("MAIN FRAME", channels=(ch_time, ch_rpm, ch_depth))

    for obj in (ch_time, ch_rpm, ch_depth, frame, origin):
        assert obj.origin_reference == 0


def test_specifying_other_origin_reference() -> None:
    df = DLISFile(storage_unit_label=make_sul())
    lf = df.add_logical_file(file_header=make_file_header())

    ch_depth = lf.add_channel(name="depth")
    ch_rpm = lf.add_channel(name="surface rpm", origin_reference=22)

    origin = lf.add_origin("DEFINING ORIGIN", origin_reference=38)

    ch_time = lf.add_channel("time", origin_reference=15)
    ch_x = lf.add_channel("x")
    frame = lf.add_frame(
        "MAIN FRAME", channels=(ch_time, ch_rpm, ch_depth, ch_x), origin_reference=10
    )

    # Items created before the logical file's defining origin (i.e., the first origin) assume the defining origin's
    # origin_reference value when it is created. This value is also assigned to items created after the defining
    # origin with no origin_reference as creation argument
    for obj in (ch_depth, ch_x, origin):
        assert obj.origin_reference == 38

    assert frame.origin_reference == 10
    assert ch_time.origin_reference == 15
    assert ch_rpm.origin_reference == 22


def test_multiple_origins() -> None:
    df = DLISFile(storage_unit_label=make_sul())
    lf = df.add_logical_file(file_header=make_file_header())

    assert lf.default_origin_reference is None
    origin1 = lf.add_origin("MAIN ORIGIN", origin_reference=23)
    assert lf.default_origin_reference == 23
    axis1 = lf.add_axis("AX1")
    ch1 = lf.add_channel("rpm")

    origin2 = lf.add_origin("ADDITIONAL ORIGIN", origin_reference=11)
    comp1 = lf.add_computation("comp1", origin_reference=origin2.origin_reference)
    comp2 = lf.add_computation("comp2")  # should have origin1 as reference
    assert lf.default_origin_reference == 23

    param1 = lf.add_parameter("PARAM1", origin_reference=11)  # origin2
    param2 = lf.add_parameter("PARAM2")  # origin1
    assert lf.default_origin_reference == 23

    for obj in (origin1, axis1, ch1, comp2, param2):
        assert obj.origin_reference == 23

    for obj in (origin2, comp1, param1):
        assert obj.origin_reference == 11



from datetime import datetime
import logging
from typing import Any
import numpy as np

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode as RepC
from dliswriter.utils.internal.struct_writer import ULONG_OFFSET
from dliswriter.utils.internal.types import number_type
from dliswriter.logical_record.core.attribute import DTimeAttribute, NumericAttribute, TextAttribute, IdentAttribute
from dliswriter.configuration import global_config


logger = logging.getLogger(__name__)


class OriginItem(EFLRItem):
    """Model an object being part of Origin EFLR."""

    parent: "OriginSet"

    def __init__(
        self,
        name: str,
        parent: "OriginSet",
        origin_reference: int,
        **kwargs: Any,
    ) -> None:
        """Initialise OriginItem.

        Args:
            name            :   Name of the OriginItem.
            parent          :   Parent OriginSet of this OriginItem.
            **kwargs        :   Values of to be set as characteristics of the OriginItem Attributes.
        """

        self.file_id = TextAttribute('file_id')
        self.file_set_name = IdentAttribute('file_set_name')
        self.file_set_number = NumericAttribute(
            'file_set_number', representation_code=RepC.UVARI, converter=self._no_reassign_file_set_number)
        self.file_number = NumericAttribute('file_number', representation_code=RepC.UVARI)
        self.file_type = IdentAttribute('file_type')
        self.product = TextAttribute('product')
        self.version = TextAttribute('version')
        self.programs = TextAttribute('programs', multivalued=True)
        self.creation_time = DTimeAttribute('creation_time')
        self.order_number = TextAttribute('order_number')
        self.descent_number = NumericAttribute('descent_number', representation_code=RepC.UNORM)
        self.run_number = NumericAttribute('run_number', representation_code=RepC.UNORM)
        self.well_id = NumericAttribute('well_id', representation_code=RepC.UNORM)
        self.well_name = TextAttribute('well_name')
        self.field_name = TextAttribute('field_name')
        self.producer_code = NumericAttribute('producer_code', representation_code=RepC.UNORM)
        self.producer_name = TextAttribute('producer_name')
        self.company = TextAttribute('company')
        self.name_space_name = IdentAttribute('name_space_name')
        self.name_space_version = NumericAttribute('name_space_version', representation_code=RepC.UVARI)

        super().__init__(
            name,
            parent=parent,
            origin_reference=origin_reference,
            **kwargs,
        )

        if self.file_set_number.value is None:
            logger.info(f"File set number for {self} not specified")
            if global_config.high_compat_mode:
                # there are poorly-understood issues related to the file set number when the file is opened in
                # Schlumberger's Log Data Composer; no such issues noticed when the file set number is 1
                # (or other low number)
                # the default file_set_number is the number of origins defined so far (for the current OriginSet)
                # (this includes the current OriginItem, so the lowest number is 1)
                n = self.parent.n_items
                logger.info(f"Setting file set number of {self} to {n}")
                self.file_set_number.value = n
            else:
                # according to the standard, file_set_number is a random integer taken from a large range
                # repr code for file_set_number is UVARI, so USHORT/UNORM/ULONG; ULONG is the largest
                max_val = np.iinfo(np.uint32).max - ULONG_OFFSET
                v = np.random.randint(1, max_val)
                logger.info(f"Setting file set number of {self} to a randomly generated number: {v}")
                self.file_set_number.value = v

        if self.creation_time.value is None:
            logger.info("Creation time ('creation_time') not specified; setting it to the current date and time")
            self.creation_time.value = datetime.now()

    def _run_checks_and_set_defaults(self) -> None:
        """Set a default field name value - if no field name has been defined."""

        if self.field_name.value is None:
            logger.debug(f"Setting field_name of {self} to the default value: 'WILDCAT'")
            self.field_name.value = 'WILDCAT'  # according to RP66

    def _no_reassign_file_set_number(self, v: number_type) -> number_type:
        """Value checker for file_set_number. Make sure the value is not reassigned."""

        if self.file_set_number.value is not None:
            raise RuntimeError("File set number should not be reassigned; to have an alternative file set number, "
                               "please pass it to the OriginItem constructor")
        return v


class OriginSet(EFLRSet):
    """Model Origin EFLR."""

    set_type = 'ORIGIN'
    logical_record_type = EFLRType.OLR
    item_type = OriginItem


OriginItem.parent_eflr_class = OriginSet



import pytest
from dlisio import dlis    # type: ignore  # untyped library

from tests.common import check_list_of_objects


@pytest.mark.parametrize(("idx", "name", "description", "object_names", "group_names"), (
        (0, "ChannelGroup", "Group of channels", ["Channel 1", "Channel 2"], []),
        (1, "ProcessGroup", "Group of processes", ["Process 1", "Prc2"], []),
        (2, "MultiGroup", "Group of groups", [], ["ChannelGroup", "ProcessGroup"]),
        (3, "Mixed-group", "Mixed objects", ["posix time", "surface rpm", "Prc2"], ["ChannelGroup", "MultiGroup"])
))
def test_group_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, description: str,
                      object_names: list[str], group_names: list[str]) -> None:
    """Test attributes of Group objects in the DLIS file."""

    g = short_dlis.groups[idx]
    assert g.name == name
    assert g.description == description
    assert g.origin == 42
    check_list_of_objects(g.objects, object_names)
    check_list_of_objects(g.groups, group_names)



import logging
from progressbar import progressbar
from typing import Optional, Sequence
from pathlib import Path

from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter.utils.internal.types import file_name_type, number_type, bytes_type
from dliswriter.logical_record.misc import StorageUnitLabel

logger = logging.getLogger(__name__)


class ByteWriter:
    """Write bytes to DLIS file."""

    def __init__(self, filename: file_name_type):
        """Initialise DLISFileWriter.

        Args:
            filename    :   Name of the file the bytes should be written to.
        """

        self._filename = filename
        self._append = False  # changes to True in first call of write_bytes
        self._total_size = 0

    @property
    def filename(self) -> file_name_type:
        """Name of the file being written."""

        return self._filename

    @property
    def total_size(self) -> int:
        """Number of bytes which have been written into the file."""

        return self._total_size

    def write_bytes(self, bts: bytes_type, size: Optional[int] = None) -> None:
        """Write (in 'wb' or 'ab' mode, as needed) the bytes into the file.

        Args:
            bts     :   Bytes to be written to the file.
            size    :   Number of bytes to be written. If not provided, it is calculated from bts.

        Note:
            For performance purposes, the provided size is not checked against the length of the bytes.
        """

        mode = 'ab' if self._append else 'wb'

        logger.debug("Writing bytes to file")
        with open(self._filename, mode) as f:
            f.write(bts)

        self._append = True  # in the future calls, append bytes to the file
        self._total_size += (size or len(bts))


class BufferedOutput:
    """Provide an automatised buffered interface for storing bytes into a file.

    Collect output bytes in a buffer of predefined size. Once buffer is full*, send the buffer contents to the
    file writer object to be stored, and prepare a new, empty buffer to collect more bytes.
    * the storing takes place when the space in the buffer is not enough to accept the provided sequence of bytes.

    Note:
        When last bytes have been passed to the buffer, it is necessary to explicitly call 'pass_bytes_to_writer'
        in order to send the bytes remaining in the buffer to the file writer.
    """

    def __init__(self, size: int, writer: ByteWriter):
        """Initialise BufferedOutput object.

        Args:
            size    :   Size of the buffer.
            writer  :   File writer object.
        """

        self._bts = bytearray(size)  #: the buffer
        self._filled_size = 0  #: how many bytes are in the current buffer
        self._buffer_size = size  #: size of the output buffer (needed when setting up a new one)

        self._writer = writer  #: file writer object

    def add_bytes(self, bts: bytes_type, size: Optional[int] = None) -> None:
        """Add bytes to the current output buffer.

        If the bytes would not fit in the current output buffer, send the currently kept bytes to the file writer,
        set up a clean output buffer, and add the new bytes there.

        Args:
            bts     :   Bytes to be added to the output buffer.
            size    :   Number of bytes to be added. If not provided, it is calculated from bts.

        Note:
            For performance purposes, the provided size is not checked against the length of the bytes.
        """

        size = size or len(bts)
        new_size = self._filled_size + size

        if new_size > self._buffer_size:
            self.pass_bytes_to_writer()  # also sets up a new output buffer
            logger.debug(f"Making new output chunk; current total output size is {self._writer.total_size}")
            new_size = size

        self._bts[self._filled_size:new_size] = bts
        self._filled_size = new_size

    def pass_bytes_to_writer(self) -> None:
        """Send the currently kept bytes to the file writer. Set up a new, empty output buffer."""

        self._writer.write_bytes(self._bts[:self._filled_size], self._filled_size)

        # set up a new output buffer
        self._bts = bytearray(self._buffer_size)
        self._filled_size = 0


class DLISWriter:
    """Create a DLIS file given data and structure information (specification of logical records)."""

    def __init__(self, filename: file_name_type, visible_record_length: int = 8192):
        """Initialise DLISFile object.

        Args:
            filename                :   Name of the file to be created. Note: at this point, no file name / directory
                                        checks (file already exists / directory write access / etc.) are performed.

            visible_record_length   :   Maximum allowed length of visible records (physical file units) in the created
                                        file. Expressed in bytes.
        """

        self._byte_writer = ByteWriter(filename)

        self._check_visible_record_length(visible_record_length)
        self._visible_record_length: int = visible_record_length  #: Maximum allowed visible record length, in bytes

        # format version is a required part of each visible record and is fixed for a given version of the standard
        self._fmt_version = RepresentationCode.USHORT.convert(255) + RepresentationCode.USHORT.convert(1)

        # flag set to True as soon as a StorageUnitLabel is written to the file (through write_storage_unit_label);
        # SUL must be the first element of the file
        self._sul_written = False

    @staticmethod
    def _check_visible_record_length(vrl: int) -> None:
        """Check the type and value of visible record length against several criteria."""

        if not isinstance(vrl, int):
            raise TypeError(f"Visible record length must be an integer; got {type(vrl)}")

        if vrl < 20:
            raise ValueError("Visible record length must be at least 20 bytes")

        if vrl > 16384:
            raise ValueError("Visible record length cannot be larger than 16384 bytes")

        if vrl % 2:
            raise ValueError("Visible record length must be an even number")

    def _make_visible_record(self, body: bytes_type, size: Optional[int] = None) -> bytes:
        """Create a visible record (physical DLIS unit) from the provided body bytes.

        Args:
            body    :   Bytes to create the visible record from.
            size    :   Number of bytes in the body. If not provided, it is calculated from the body object.

        Returns:
            Created visible record (provided body bytes preceded by header bytes) as a bytes object.

        Note:
            For performance purposes, the provided size is not checked against the length of the body bytes.
        """

        if size is None:
            size = len(body)

        size += 4  # 4 header bytes will be added

        if size > self._visible_record_length:
            raise ValueError(f"VR length is too large; got {size}, max is {self._visible_record_length}")

        return RepresentationCode.UNORM.convert(size) + self._fmt_version + body

    def _check_output_chunk_size(self, output_chunk_size: number_type) -> None:
        """Check output chunk size type (integer or float with zero decimal part) and value (>= max VR length)."""

        if not isinstance(output_chunk_size, (int, float)):
            raise TypeError(f"Output chunk size must be a number; got {type(output_chunk_size)}")

        if output_chunk_size % 1:
            raise ValueError(f"Output chunk size must be an integer; got {output_chunk_size}")

        if output_chunk_size < self._visible_record_length:
            raise ValueError(f"Output chunk size cannot be smaller than max visible record length "
                             f"(= {self._visible_record_length}); got {output_chunk_size}")

    def write_storage_unit_label(self, sul: StorageUnitLabel) -> None:
        """Convert a Storage Unit Label to bytes and write them into the file.

        Unlike other, 'proper' logical records, SUL should not be wrapped in a visible record structure,
        but instead written to the file as-is.
        """

        logger.info("Writing Storage Unit Label bytes to the file")
        self._byte_writer.write_bytes(sul.represent_as_bytes().bts)
        self._sul_written = True

    def write_logical_records(self, logical_records: Sequence, output_chunk_size: Optional[number_type]) -> None:
        """Write the provided logical records to the file.

        Note: write_storage_unit_label MUST be called BEFORE calling this method.
        Otherwise, a RuntimeError is raised.

        Args:
            logical_records     :   Logical records to become part of the file.
            output_chunk_size   :   Size of the buffers accumulating file bytes before file write action is called.
        """

        if not self._sul_written:
            raise RuntimeError("Storage Unit Label absent from the file; "
                               "add it calling DLISWriter.write_storage_unit_label")

        # prepare BufferedOutput object - temporarily keep added bytes, store them in the file when buffer is full
        output_chunk_size = output_chunk_size or 2 ** 32
        self._check_output_chunk_size(output_chunk_size)
        logger.debug(f"Output file will be produced in chunks of max size {output_chunk_size} bytes")
        output = BufferedOutput(int(output_chunk_size), self._byte_writer)

        # max allowed size of an LR segment body; 4 bytes reserved for VR header and another 4 for LR segment header
        max_lr_segment_size = self._visible_record_length - 8

        # loop through the logical records, transform them and write them to the file
        logger.info("Creating & writing visible records of the DLIS...")
        for lr in progressbar(logical_records, max_value=len(logical_records)):
            # represent a logical record as bytes; split it segments as needed
            for segment, segment_size in lr.represent_as_bytes().make_segments(max_lr_segment_size):
                # wrap each segment's bytes in a separate visible record and write the VR to the file
                output.add_bytes(self._make_visible_record(segment, segment_size))
        output.pass_bytes_to_writer()  # pass the remaining bytes kept in the output buffer (not full atm) to the writer

        # summarise
        logger.info(f'{len(logical_records)} written to DLIS file at {Path(self._byte_writer.filename).resolve()}')
        logger.info(f"Total file size is {self._byte_writer.total_size} bytes")



import pytest

from dliswriter.logical_record.eflr_types.group import GroupSet, GroupItem
from dliswriter.logical_record.eflr_types.channel import ChannelItem
from dliswriter.logical_record.eflr_types.process import ProcessItem
from dliswriter.logical_record.core.eflr import EFLRItem
from dliswriter import high_compatibility_mode_decorator


@pytest.mark.parametrize(("name", "description", "object_type", "object_class", "object_names", "group_names"), (
        ("ChannelGroup", "Group of channels", "CHANNEL", ChannelItem, ['channel1', 'channel2'], []),
        ("ProcessGroup", "Group of processes", "PROCESS", ProcessItem, ["process1", "process2"], []),
        ("MultiGroup", "Group of groups", "GROUP", None, [], ["channel_group", "process_group"])
))
def test_group_params(name: str, description: str, object_type: str,
                      object_class: type[EFLRItem], object_names: list[str], group_names: list[str],
                      request: pytest.FixtureRequest) -> None:
    """Test creating GroupItem."""

    g = GroupItem(
        name,
        description=description,
        object_type=object_type,
        object_list=[request.getfixturevalue(v) for v in object_names],
        group_list=[request.getfixturevalue(v) for v in group_names],
        parent=GroupSet()
    )

    assert g.name == name
    assert g.description.value == description
    assert g.object_type.value == object_type

    assert len(g.object_list.value) == len(object_names)
    for i, obj in enumerate(g.object_list.value):
        assert isinstance(obj, object_class)
        assert obj is request.getfixturevalue(object_names[i])

    assert len(g.group_list.value) == len(group_names)
    for i, grp in enumerate(g.group_list.value):
        assert isinstance(grp, GroupItem)
        assert grp is request.getfixturevalue(group_names[i])

    assert isinstance(g.parent, GroupSet)
    assert g.parent.set_name is None


@pytest.mark.parametrize("name", ("GROUP-1", "GROUP_OF_CHANNELS", "12-CHANNEL_GROUP"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    GroupItem(name, parent=GroupSet())


@pytest.mark.parametrize("name", ("Group 1", "Group#12", "A.5"))
@high_compatibility_mode_decorator
def test_name_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        GroupItem(name, parent=GroupSet())


@pytest.mark.parametrize("ot", ("GROUP", "CHANNEL", "FRAME"))
@high_compatibility_mode_decorator
def test_object_type_compatible(ot: str) -> None:
    GroupItem("G1", object_type=ot, parent=GroupSet())


@pytest.mark.parametrize("ot", ("Channel", "CALIBRATION MEASUREMENT", "MEASUREMENTS#5"))
@high_compatibility_mode_decorator
def test_object_type_not_compatible(ot: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        GroupItem("G12", object_type=ot, parent=GroupSet())



from dliswriter.utils.internal.validator_enum import ValidatorEnum


class Unit(ValidatorEnum):
    """All units explicitly allowed by the RP66 standard."""

    AMPERE = "A"
    KELVIN = "K"
    CANDELA = "cd"
    API_GRAVITY = "dAPI"
    DECIBEL = "dB"
    API_GAMMA_RAY = "gAPI"
    KILOGRAM = "kg"
    METER = "m"
    MOLE = "mol"
    API_NEUTRON = "nAPI"
    RADIAN = "rad"
    SECOND = "s"
    STERADIAN = "sr"
    BRITISH_THERMAL_UNIT = "Btu"
    COULOMB = "C"
    DARCY = "D"
    GIGAPASCAL = "GPa"
    GAL = "Gal"
    HERTZ = "Hz"
    JOULE = "J"
    LITER = "L"
    MEGAHERTZ = "MHz"
    MEGAPASCAL = "MPa"
    MEGAELECTRONVOLT = "MeV"
    MEGAGRAM = "Mg"
    MEGA_POUNDS_PER_SQUARE_METER = "Mpsi"
    NEWTON = "N"
    OERSTED = "Oe"
    POISE = "P"
    PASCAL = "Pa"
    SIEMENS = "S"
    TESLA = "T"
    VOLT = "V"
    WATT = "W"
    WEBER = "Wb"
    ANNUM = "a"
    ACRE = "acre"
    STD_ATMOSPHERE = "atm"
    BARN = "b"
    BAR = "bar"
    BARREL = "bbl"
    REVOLUTION_CYCLE = "c"
    CENTIPOISE = "cP"
    CALORIE = "cal"
    CENTIMETER = "cm"
    CAPTURE_UNIT = "cu"
    DAY = "d"
    DECANEWTON = "daN"
    DEGREE_ANGLE = "deg"
    DEGREE_CELSIUS = "degC"
    DEGREE_FAHRENHEIT = "degF"
    DECIMETER = "dm"
    ELECTRONVOLT = "eV"
    FEMTOCOULOMB = "fC"
    FOOT = "ft"
    GRAM = "g"
    GALLON = "gal"
    HOUR = "h"
    INCH = "in"
    KILOHERTZ = "kHz"
    KILOPASCAL = "kPa"
    KILOVOLT = "kV"
    KILOELECTRONVOLT = "keV"
    KILOGRAM_FORCE = "kgf"
    KILOMETER = "km"
    POUND_FORCE = "lbf"
    POUND_MASS = "lbm"
    MILLIAMPERE = "mA"
    MILLICOULOMB = "mC"
    MILLIDARCY = "mD"
    MILLIGAL = "mGal"
    MILLILITER = "mL"
    MILLISIEMENS = "mS"
    MILLITESLA = "mT"
    MILLIVOLT = "mV"
    MILLIWAT = "mW"
    MILLIGRAM = "mg"
    MINUTE = "min"
    MILLIMETER = "mm"
    MILLIOHM = "mohm"
    MILLISECOND = "ms"
    NANOCOULOMB = "nC"
    NANOWATT = "nW"
    NANOSECOND = "ns"
    OHM = "ohm"
    PICCOCOULOMB = "pC"
    PICOPASCAL = "pPa"
    PART_PER_TEN_THOUSAND = "ppdk"
    PART_PER_THOUSAND = "ppk"
    PART_PER_MILLION = "ppm"
    POUND_PER_SQUARE_INCH = "psi"
    POROSITY_UNIT = "pu"
    METRIC_TON = "t"
    US_SHORT_TON = "ton"
    MICROAMPERE = "uA"
    MICROCOULOMB = "uC"
    MICROPASCAL = "uPa"
    MICROVOLT = "uV"
    MICROMETER = "um"
    MICROOHM = "uohm"
    MICRO_POUND_PER_SQUARE_INCH = "upsi"
    MICROSECOND = "us"


class Property(ValidatorEnum):
    """Allowed values for elements of the 'properties' attribute of Channel, Computation, and Process"""

    AVERAGED = 'AVERAGED'
    CALIBRATED = 'CALIBRATED'
    CHANGED_INDEX = 'CHANGED-INDEX'
    COMPUTED = 'COMPUTED'
    DEPTH_MATCHED = 'DEPTH-MATCHED'
    DERIVED = 'DERIVED'
    FILTERED = 'FILTERED'
    HOLE_SIZE_CORRECTED = 'HOLE-SIZE-CORRECTED'
    INCLINOMETRY_CORRECTED = 'INCLINOMETRY-CORRECTD'
    LITHOLOGY_CORRECTED = 'LITHOLOGY-CORRECTED'
    LOCAL_COMPUTATION = 'LOCAL-COMPUTATION'
    LOCALLY_DEFINED = 'LOCALLY-DEFINED'
    MODELLED = 'MODELLED'
    MUDCAKE_CORRECTED = 'MUDCAKE-CORRECTED'
    NORMALIZED = 'NORMALIZED'
    OVERSAMPLED = 'OVER-SAMPLED'
    PATCHED = 'PATCHED'
    PRESSURE_CONTROLLED = 'PRESSURE-CORRECTED'
    RESAMPLED = 'RE-SAMPLED'
    SALINITY_CONTROLLED = 'SALINITY-CORRECTED'
    SAMPLED_DOWNWARD = 'SAMPLED-DOWNWARD'
    SAMPLED_UPWARD = 'SAMPLED-UPWARD'
    SPEED_CORRECTED = 'SPEED-CORRECTED'
    SPLICED = 'SPLICED'
    SQUARED = 'SQUARED'
    STACKED = 'STACKED'
    STD = 'STANDARD-DEVIATION'
    STANDOFF_CORRECTED = 'STANDOFF-CORRECTED'
    TEMPERATURE_CORRECTED = 'TEMPERATURE-CORRECTED'
    UNDERSAMPLED = 'UNDER-SAMPLED'


class CalibrationMeasurementPhase(ValidatorEnum):
    """Allowed values for 'phase' of a CalibrationMeasurementItem."""

    AFTER = 'AFTER'
    BEFORE = 'BEFORE'
    MASTER = 'MASTER'


class EquipmentType(ValidatorEnum):
    """Options for the 'type' ('_type') attribute of Equipment, allowed by the standard"""

    ADAPTER = "Adapter"
    BOARD = "Board"
    BOTTOM_NOSE = "Bottom-Nose"
    BRIDLE = "Bridle"
    CABLE = "Cable"
    CALIBRATOR = "Calibrator"
    CARTRIDGE = "Cartridge"
    CENTRALIZER = "Centralizer"
    CHAMBER = "Chamber"
    CUSHION = "Cushion"
    DEPTH_DEVICE = "Depth-Device"
    DISPLAY = "Display"
    DRAWER = "Drawer"
    EXCENTRALIZER = "Excentralizer"
    EXPLOSIVE_SOURCE = "Explosive-Source"
    FLASK = "Flask"
    GEOPHONE = "Geophone"
    GUN = "Gun"
    HEAD = "Head"
    HOUSING = "Housing"
    JIG = "Jig"
    JOINT = "Joint"
    NUCLEAR_DETECTOR = "Nuclear-Detector"
    PACKER = "Packer"
    PAD = "Pad"
    PANE = "Pane"
    POSITIONING = "Positioning"
    PRINTER = "Printer"
    RADIOACTIVE_SOURCE = "Radioactive-Source"
    SHIELD = "Shield"
    SIMULATOR = "Simulator"
    SKID = "Skid"
    SONDE = "Sonde"
    SPACER = "Spacer"
    STANDOFF = "Standoff"
    SYSTEM = "System"
    TOOL = "Tool"
    TOOL_MODULE = "Tool-Module"
    TRANSDUCER = "Transducer"
    VIBRATION_SOURCE = "Vibration-Source"


class EquipmentLocation(ValidatorEnum):
    """Options for the 'location' attribute of Equipment, allowed by the standard."""

    LOGGING_SYSTEM = 'Logging-System'
    REMOTE = 'Remote'
    RIG = 'Rig'
    WELL = 'Well'


class FrameIndexType(ValidatorEnum):
    """Values for frame index type allowed by the standard."""

    ANGULAR_DRIFT = 'ANGULAR-DRIFT'
    BOREHOLE_DEPTH = 'BOREHOLE-DEPTH'
    NON_STANDARD = 'NON-STANDARD'
    RADIAL_DRIFT = 'RADIAL-DRIFT'
    VERTICAL_DEPTH = 'VERTICAL-DEPTH'


class ProcessStatus(ValidatorEnum):
    """Allowed values of the 'status' Attribute of Process."""

    COMPLETE = 'COMPLETE'
    ABORTED = 'ABORTED'
    IN_PROGRESS = 'IN-PROGRESS'


class ZoneDomain(ValidatorEnum):
    """Allowed values for 'domain' Attribute of Zone.

    Comments are quotes from RP66; 'Zone interval is'...:"""

    BOREHOLE_DEPTH = 'BOREHOLE-DEPTH'   # 'along the borehole'
    TIME = 'TIME'                       # 'elapsed time'
    VERTICAL_DEPTH = 'VERTICAL-DEPTH'   # 'depth along the Vertical Generatrix'



import pytest

from dliswriter.logical_record.misc import StorageUnitLabel
from dliswriter import high_compatibility_mode_decorator


@pytest.mark.parametrize(("name", "sequence_number"), (("SUL-1", 11), ("Default storage", 0)))
def test_sul_creation(name: str, sequence_number: int) -> None:
    """Test creating StorageUnitLabel."""

    sul = StorageUnitLabel(name, sequence_number=sequence_number)

    assert sul.set_identifier == name
    assert sul.sequence_number == sequence_number


@pytest.mark.parametrize("name", ("SUL-1", "UNIT1", "123-11"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    StorageUnitLabel(name)


@pytest.mark.parametrize("name", ("STORAGE UNIT LABEL", "SUL#12", "12.3"))
@high_compatibility_mode_decorator
def test_name_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        StorageUnitLabel(name)



import logging
import pytest
import numpy as np
from typing import Optional

from dliswriter.logical_record.eflr_types.frame import FrameSet, FrameItem
from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter import DLISFile, high_compatibility_mode_decorator, enums


def test_frame_creation() -> None:
    """Test creating d FrameObject."""

    frame = FrameItem(
        "MAIN-FRAME",
        **{
            "index_type": "BOREHOLE-DEPTH",
            "encrypted": 1,
            "description": "The main frame",
            "spacing": {"value": 0.2, "units": "m"},
        },
        parent=FrameSet(),
    )

    assert frame.name == "MAIN-FRAME"
    assert frame.index_type.value == "BOREHOLE-DEPTH"
    assert frame.encrypted.value == 1
    assert frame.description.value == "The main frame"

    assert frame.spacing.value == 0.2
    assert frame.spacing.units == "m"
    assert frame.spacing.representation_code is RepresentationCode.FDOUBL

    assert isinstance(frame.parent, FrameSet)
    assert frame.parent.set_name is None


@pytest.mark.parametrize(
    "channel_names", (("Channel 1", "Channel 3", "Channel 2"), ("some_channel",))
)
def test_creation_with_channels(channel_names: tuple[str], channels: dict) -> None:
    """Test creating a FrameObject with specified channels."""

    frame = FrameItem(
        "Some frame", channels=[channels[k] for k in channel_names], parent=FrameSet()
    )

    assert frame.channels.value is not None
    assert len(frame.channels.value) == len(channel_names)

    for i, cn in enumerate(channel_names):
        assert frame.channels.value[i].name == cn


def _prepare_file_uneven_spacing() -> DLISFile:
    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("ORIGIN")
    ch1 = lf.add_channel("INDEX", data=np.arange(10) + np.random.rand(10))
    ch2 = lf.add_channel("X", units="m", data=np.random.rand(10, 10))
    lf.add_frame(
        "MAIN", channels=(ch1, ch2), index_type=enums.FrameIndexType.NON_STANDARD
    )
    return df


def test_spacing_not_even(caplog: pytest.LogCaptureFixture) -> None:
    df = _prepare_file_uneven_spacing()

    with caplog.at_level(logging.WARNING, logger="dliswriter"):
        df.generate_logical_records(chunk_size=None)
        assert "Spacing of the index channel" in caplog.text


@high_compatibility_mode_decorator
def test_spacing_not_even_high_compat_mode() -> None:
    df = _prepare_file_uneven_spacing()

    with pytest.raises(
        RuntimeError, match="Spacing of the index channel .* is not uniform.*"
    ):
        df.generate_logical_records(chunk_size=None)


def _prepare_file_first_channel_2d(
    index_type: Optional[enums.FrameIndexType] = None,
) -> DLISFile:
    df = DLISFile()
    lf = df.add_logical_file()
    lf.add_origin("DEFINING ORIGIN")
    ch1 = lf.add_channel("INDEX", data=np.arange(start=0, stop=24, dtype=np.uint32).reshape(12, 2))
    ch2 = lf.add_channel("X", data=np.random.rand(12))
    ch3 = lf.add_channel("Y", data=np.random.randint(0, 10, size=(12, 10), dtype=np.int32))
    lf.add_frame("F1", channels=(ch1, ch2, ch3), index_type=index_type)
    return df


def test_index_channel_not_1d() -> None:
    df = _prepare_file_first_channel_2d(index_type=enums.FrameIndexType.NON_STANDARD)
    with pytest.raises(
        RuntimeError, match="Index channel's data must be 1-dimensional.*"
    ):
        df.generate_logical_records(chunk_size=None)


def test_first_channel_not_1d_but_no_index() -> None:
    df = _prepare_file_first_channel_2d(index_type=None)
    df.generate_logical_records(chunk_size=None)  # goes through without error



from dliswriter.logical_record.eflr_types.long_name import LongNameSet, LongNameItem


def test_from_config() -> None:
    """Test creating LongNameObject."""

    t = 'SOME ASCII TEXT'

    w = LongNameItem(
        "some long name",
        general_modifier=[t],
        quantity=t,
        quantity_modifier=[t],
        altered_form=t,
        entity=t,
        entity_modifier=[t],
        entity_number=t,
        entity_part=t,
        entity_part_number=t,
        generic_source=t,
        source_part=[t],
        source_part_number=[t],
        conditions=[t],
        standard_symbol=t,
        private_symbol=t,
        parent=LongNameSet()
    )

    assert w.general_modifier.value == [t]
    assert w.quantity.value == t
    assert w.quantity_modifier.value == [t]
    assert w.altered_form.value == t
    assert w.entity.value == t
    assert w.entity_modifier.value == [t]
    assert w.entity_number.value == t
    assert w.entity_part.value == t
    assert w.entity_part_number.value == t
    assert w.generic_source.value == t
    assert w.source_part.value == [t]
    assert w.source_part_number.value == [t]
    assert w.conditions.value == [t]
    assert w.standard_symbol.value == t
    assert w.private_symbol.value == t

    assert isinstance(w.parent, LongNameSet)
    assert w.parent.set_name is None



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode as RepC
from dliswriter.logical_record.core.attribute import EFLRAttribute, TextAttribute, IdentAttribute
from dliswriter.utils.internal.value_checkers import validate_string


logger = logging.getLogger(__name__)


class GroupItem(EFLRItem):
    """Model an object being part of Group EFLR."""

    parent: "GroupSet"

    def __init__(self, name: str, parent: "GroupSet", **kwargs: Any) -> None:
        """Initialise GroupItem.

        Args:
            name        :   Name of the GroupItem.
            parent      :   Group EFLRSet instance this GroupItem belongs to.
            **kwargs    :   Values of to be set as characteristics of the GroupItem Attributes.
        """

        self.description = TextAttribute('description')
        self.object_type = IdentAttribute('object_type', converter=validate_string)
        self.object_list = EFLRAttribute('object_list', multivalued=True, representation_code=RepC.OBJREF)
        self.group_list = EFLRAttribute('group_list', object_class=GroupSet, multivalued=True)

        super().__init__(name, parent=parent, **kwargs)


class GroupSet(EFLRSet):
    """Model Group EFLR."""

    set_type = 'GROUP'
    logical_record_type = EFLRType.STATIC
    item_type = GroupItem


GroupItem.parent_eflr_class = GroupSet



from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import Attribute, NumericAttribute, IdentAttribute
from dliswriter.utils.internal.value_checkers import validate_string, convert_maybe_numeric


class AxisItem(EFLRItem):
    """Model an object being part of Axis EFLR."""

    parent: "AxisSet"

    def __init__(self, name: str, parent: "AxisSet", **kwargs: Any) -> None:
        """Initialise AxisItem.

        Args:
            name        :   Name of the AxisItem.
            parent      :   Parent AxisSet of this AxisItem.
            **kwargs    :   Values of to be set as characteristics of the AxisItem Attributes.
        """

        self.axis_id = IdentAttribute('axis_id', converter=validate_string)
        self.coordinates = Attribute('coordinates', multivalued=True, converter=convert_maybe_numeric)
        self.spacing = NumericAttribute('spacing')

        super().__init__(name, parent=parent, **kwargs)


class AxisSet(EFLRSet):
    """Model Axis EFLR."""

    set_type = 'AXIS'
    logical_record_type = EFLRType.AXIS
    item_type = AxisItem


AxisItem.parent_eflr_class = AxisSet



import pytest
from dlisio import dlis    # type: ignore  # untyped library
from typing import Union

from tests.common import check_list_of_objects


def test_parameters(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of parameters in the DLIS file matches the expected one."""

    params = short_dlis.parameters
    assert len(params) == 3


@pytest.mark.parametrize(("idx", "name", "long_name", "values", "zones"), (
        (0, "Param-1", "LATLONG-GPS", ["40deg 23' 42.8676'' N", "40deg 23' 42.8676'' N"], ["Zone-1", "Zone-3"]),
        (1, "Param-2", "LATLONG", [[40.395241, 27.792471, 21.23131213], [21, 23, 24]], ["Zone-2", "Zone-4"]),
        (2, "Param-3", None, [12.5], [])
))
def test_parameters_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, long_name: Union[str, None],
                           values: list, zones: list[str]) -> None:
    """Check attributes of DLIS Parameter objects."""

    if long_name is None:
        long_name = short_dlis.longnames[0]

    param = short_dlis.parameters[idx]
    assert param.name == name
    assert param.long_name == long_name
    assert param.values.tolist() == values
    assert param.origin == 42

    check_list_of_objects(param.zones, zones)



from dlisio import dlis    # type: ignore  # untyped library


from tests.common import N_COLS, select_channel


def test_channel_properties(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of channels in the new DLIS file."""

    for name in ('posix time', 'surface rpm'):
        chan = select_channel(short_dlis, name)
        assert chan.name == name
        assert chan.element_limit == [1]
        assert chan.dimension == [1]
        assert chan.origin == 42

    for name in ('amplitude', 'radius', 'radius_pooh'):
        chan = select_channel(short_dlis, name)
        assert chan.name == name
        assert chan.element_limit == [N_COLS]
        assert chan.dimension == [N_COLS]
        assert chan.origin == 42

    assert short_dlis.object("CHANNEL", 'amplitude').units is None
    assert short_dlis.object("CHANNEL", 'radius').units == "in"
    assert short_dlis.object("CHANNEL", 'radius_pooh').units == "m"


def test_channel_long_name(short_dlis: dlis.file.LogicalFile) -> None:
    """Check long name channels in the new DLIS file."""

    ln = short_dlis.longnames[0]

    for name in ('Channel 2', 'surface rpm'):
        chan = select_channel(short_dlis, name)
        assert chan.long_name is ln

    chan = select_channel(short_dlis, "Some Channel")
    assert chan.long_name == "Some not so very long channel name"

    chan = select_channel(short_dlis, "channel_x")
    assert chan.long_name == "Channel not added to the frame"


def test_channel_not_in_frame(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that channel which was not added to frame is still in the file."""

    name = 'channel_x'
    select_channel(short_dlis, name)  # if no error - channel is found in the file
    assert not any(c.name == name for c in short_dlis.frames[0].channels)



from dlisio import dlis    # type: ignore  # untyped library


def test_frame(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attributes of Frame in the new DLIS file."""

    assert len(short_dlis.frames) == 1

    frame = short_dlis.frames[0]

    assert frame.name == "MAIN FRAME"
    assert frame.index_type == "TIME"
    assert frame.origin == 42



from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import TextAttribute


class CommentItem(EFLRItem):
    """Model an object being part of Comment EFLR."""

    parent: "CommentSet"

    def __init__(self, name: str, parent: "CommentSet", **kwargs: Any) -> None:
        """Initialise CommentItem.

        Args:
            name        :   Name of the CommentItem.
            parent      :   Parent CommentSet of this CommentItem.
            **kwargs    :   Values of to be set as characteristics of the CommentItem Attributes.
        """

        self.text = TextAttribute('text', multivalued=True)

        super().__init__(name, parent=parent, **kwargs)


class CommentSet(EFLRSet):
    """Model Comment EFLR."""

    set_type = 'COMMENT'
    logical_record_type = EFLRType.SCRIPT
    item_type = CommentItem


CommentItem.parent_eflr_class = CommentSet



"""This is an example of how to create a DLIS file with multiple frames with equivalent channels - i.e. the same channel
names used in different frames (but referring to separate datasets).

The procedure is the same as for standard DLIS files.
When inspecting the objects, one might notice that:
    - copy_number of the channels is different - 0 for channels of frame 1, 1 for channels of frame 2
    - dataset_name attributes of the corresponding channels are also different, so that datasets are not overwritten.
"""

import numpy as np
import logging

from dliswriter import DLISFile, EFLRItem, enums, eflr_types

from utils import install_colored_logger


# colored logs output
install_colored_logger(logging.getLogger('dliswriter'))


# create DLISFile instance; optionally, pass arguments for creating file header & storage unit label
df = DLISFile(sul_sequence_number=2)

lf = df.add_logical_file(
    fh_sequence_number=2,
    fh_id="MAIN FILE",
    fh_identifier="X"
)

# add origin - required item
lf.add_origin(
    "DEFAULT ORIGIN",
    file_set_number=80,
    company="XXX"
)


# define frame 1
n_rows_1 = 100
ch_depth_1 = lf.add_channel(
    'DEPTH',
    data=np.arange(n_rows_1),
    units=enums.Unit.METER
)
ch_rpm_1 = lf.add_channel(
    "RPM",
    data=10 * np.random.rand(n_rows_1)
)
ch_amp_1 = lf.add_channel(
    "AMPLITUDE",
    data=np.random.rand(n_rows_1, 10)
)
frame1 = lf.add_frame(
    "FRAME1",
    channels=(ch_depth_1, ch_rpm_1, ch_amp_1),
    index_type=enums.FrameIndexType.BOREHOLE_DEPTH
)


# define frame 2
n_rows_2 = 200
ch_depth_2 = lf.add_channel(
    'DEPTH',
    data=np.arange(n_rows_2),
    units=enums.Unit.METER
)
ch_rpm_2 = lf.add_channel(
    "RPM",
    data=(np.arange(n_rows_2) % 10).astype(np.int32)
)
ch_amp_2 = lf.add_channel(
    "AMPLITUDE",
    data=np.arange(n_rows_2 * 5).reshape(n_rows_2, 5) % 6
)
frame2 = lf.add_frame(
    "FRAME2",
    channels=(ch_depth_2, ch_rpm_2, ch_amp_2),
    index_type=enums.FrameIndexType.BOREHOLE_DEPTH
)


# write the file
df.write('./tmp.DLIS')


# print the copy_number and dataset_name info of the objects for inspection
def describe_item(o: EFLRItem) -> None:
    s = f"{o}: \n\tcopy_number = {o.copy_number}"

    if isinstance(o, eflr_types.ChannelItem):
        s += f"\n\tdataset_name = {o.dataset_name}"

    print(s)


for eflr_item in (frame1, ch_depth_1, ch_rpm_1, ch_amp_1, frame2, ch_depth_2, ch_rpm_2, ch_amp_2):
    describe_item(eflr_item)



"""From RP66 v1:
'Each Logical File begins with a File Header Logical Record that serves as an identifying label for the Logical File.'

Unless otherwise specified, all quotes in this file come from the RP66 v1 standard specification.
"""

from typing import Any

from dliswriter.utils.internal.converters import get_ascii_bytes
from dliswriter.utils.internal.struct_writer import write_struct_ascii
from dliswriter.utils.internal.internal_enums import RepresentationCode, EFLRType
from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.value_checkers import validate_string


def pack_ushort(v: int) -> bytes:
    return RepresentationCode.USHORT.convert(v)


class FileHeaderItem(EFLRItem):
    """Model an object being part of FileHeader EFLR."""

    parent: "FileHeaderSet"

    header_id_length_limit = 65             #: max length of the file header name
    max_sequence_number = int(1e10 - 1)     #: max value for sequence number; largest 10-digit integer

    def __init__(self, header_id: str, parent: "FileHeaderSet", sequence_number: int = 1,
                 identifier: str = '0') -> None:
        """Initialise FileHeaderItem.

        Args:
            header_id       :   '[A] descriptive identification of the Logical File'.
            parent          :   Parent FileHeaderSet of this FileHeaderItem.
            sequence_number :   'The ASCII representation of a positive integer that indicates the sequential position
                                of the Logical File in a Storage Set'.
                                Must be a positive integer whose ASCII representation does not exceed 10 characters.
            identifier      :   '[A] single arbitrary character', identifying the File Header.
        """

        if not isinstance(header_id, str):
            raise TypeError(f"'header_id' should be a str; got {type(header_id)}")
        if len(header_id) > self.header_id_length_limit:
            raise ValueError(f"'header_id' length should not exceed {self.header_id_length_limit} characters")

        if not isinstance(sequence_number, int):
            raise TypeError(f"'sequence_number' should be an integer; got {type(sequence_number)}: {sequence_number}")
        if not 0 < sequence_number <= self.max_sequence_number:
            raise ValueError(f"Sequence number must be a positive integer not larger than {self.max_sequence_number}; "
                             f"got {sequence_number}")

        if not isinstance(identifier, str):
            raise TypeError(f"'identifier' should be a str; got {type(identifier)}: {identifier}")
        if len(identifier) != 1:
            raise ValueError(f"'identifier' should be a single character; got '{identifier}'")

        self.header_id = validate_string(header_id)
        self.sequence_number = sequence_number

        super().__init__(name=identifier, parent=parent)

    def __repr__(self) -> str:
        return (f"{self.__class__.__name__}(header_id={self.header_id}, sequence_number={self.sequence_number}, "
                f"parent=FileHeaderSet())")

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, type(self)):
            return False

        if not self.sequence_number == other.sequence_number:
            return False

        if not self.header_id == other.header_id:
            return False

        if not self.parent.set_name == other.parent.set_name:
            return False

        return True

    def _make_attrs_bytes(self) -> bytes:
        """Create bytes describing the values of attributes of FIleHeaderItem."""

        bts = b''

        bts += pack_ushort(int('00100001', 2))
        bts += pack_ushort(10)
        bts += get_ascii_bytes(str(self.sequence_number), 10, justify_left=False)
        bts += pack_ushort(int('00100001', 2))
        bts += pack_ushort(65)
        bts += get_ascii_bytes(self.header_id, 65, justify_left=True)

        return bts


class FileHeaderSet(EFLRSet):
    """Model FileHeader EFLR."""

    set_type = 'FILE-HEADER'
    logical_record_type = EFLRType.FHLR
    item_type = FileHeaderItem

    def __init__(self) -> None:
        super().__init__(set_name=None)

    def _make_template_bytes(self) -> bytes:
        """Create bytes describing the template - kinds of attributes to be found in the FileHeader EFLR."""

        bts = b''

        bts += pack_ushort(int('00110100', 2))
        bts += write_struct_ascii('SEQUENCE-NUMBER')
        bts += pack_ushort(20)

        bts += pack_ushort(int('00110100', 2))
        bts += write_struct_ascii('ID')
        bts += pack_ushort(20)

        return bts


FileHeaderItem.parent_eflr_class = FileHeaderSet



"""From RP66 v1:

`Zone Objects specify single intervals in depth or time. Zone Objects are useful for associating other Objects
or values with specific regions of a well or with specific time intervals.`
"""

from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode
from dliswriter.utils.enums import ZoneDomain
from dliswriter.logical_record.core.attribute import IdentAttribute, DTimeAttribute, TextAttribute


class ZoneItem(EFLRItem):
    """Model an object being part of Zone EFLR."""

    parent: "ZoneSet"

    def __init__(self, name: str, parent: "ZoneSet", **kwargs: Any) -> None:
        """Initialise ZoneItem.

        Args:
            name        :   Name of the ZoneItem.
            parent      :   Parent ZoneSet of this ZoneItem.
            **kwargs    :   Values of to be set as characteristics of the ZoneItem Attributes.
        """

        self.description = TextAttribute('description')
        self.domain = IdentAttribute(
            'domain', converter=ZoneDomain.make_converter('domains'))
        self.maximum = DTimeAttribute('maximum', allow_float=True)
        self.minimum = DTimeAttribute('minimum', allow_float=True)

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Check maximum and minimum vs domain before writing the object."""

        domain = self.domain.value

        if domain is None:
            return

        rcs = [self.maximum.representation_code, self.minimum.representation_code]
        rcs = [rc for rc in rcs if rc is not None]
        if not rcs:
            return

        codes_are_dtime = [rc is RepresentationCode.DTIME for rc in rcs]

        if domain == 'TIME':
            if any(codes_are_dtime) and not all(codes_are_dtime):
                raise RuntimeError(f"{self}: either both or none of Zone's maximum and minimum should be of datetime "
                                   f"type; got {self.minimum.value} and {self.maximum.value}")

        else:
            if any(codes_are_dtime):
                raise RuntimeError(f"{self}: domain is '{domain}', so only float values for minimum and maximum "
                                   f"are allowed")


class ZoneSet(EFLRSet):
    """Model Zone EFLR."""

    set_type = 'ZONE'
    logical_record_type = EFLRType.STATIC
    item_type = ZoneItem


ZoneItem.parent_eflr_class = ZoneSet



from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import IdentAttribute, DTimeAttribute, NumericAttribute, TextAttribute


class MessageItem(EFLRItem):
    """Model an object being part of Message EFLR."""

    parent: "MessageSet"

    def __init__(self, name: str, parent: "MessageSet", **kwargs: Any) -> None:
        """Initialise MessageItem.

        Args:
            name        :   Name of the MessageItem.
            parent      :   Parent MessageSet of this MessageItem.
            **kwargs    :   Values of to be set as characteristics of the MessageItem Attributes.
        """

        self._type = IdentAttribute('_type')
        self.time = DTimeAttribute('time', allow_float=True)
        self.borehole_drift = NumericAttribute('borehole_drift')
        self.vertical_depth = NumericAttribute('vertical_depth')
        self.radial_drift = NumericAttribute('radial_drift')
        self.angular_drift = NumericAttribute('angular_drift')
        self.text = TextAttribute('text', multivalued=True)

        super().__init__(name, parent=parent, **kwargs)


class MessageSet(EFLRSet):
    """Model Message EFLR."""

    set_type = 'MESSAGE'
    logical_record_type = EFLRType.SCRIPT
    item_type = MessageItem


MessageItem.parent_eflr_class = MessageSet



from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import IdentAttribute, TextAttribute


class NoFormatItem(EFLRItem):
    """Model an object being part of NoFormat EFLR."""

    parent: "NoFormatSet"

    def __init__(self, name: str, parent: "NoFormatSet", **kwargs: Any) -> None:
        """Initialise NoFormatItem.

        Args:
            name        :   Name of the NoFormatItem.
            parent      :   Parent NoFormatSet of this NoFormatItem.
            **kwargs    :   Values of to be set as characteristics of the NoFormatItem Attributes.
        """

        self.consumer_name = IdentAttribute('consumer_name')
        self.description = TextAttribute('description')

        super().__init__(name, parent=parent, **kwargs)


class NoFormatSet(EFLRSet):
    """Model NoFormat EFLR."""

    set_type = 'NO-FORMAT'
    logical_record_type = EFLRType.UDI
    item_type = NoFormatItem


NoFormatItem.parent_eflr_class = NoFormatSet



import pytest

from dliswriter import high_compatibility_mode_decorator
from dliswriter.logical_record.eflr_types import CalibrationCoefficientItem, CalibrationCoefficientSet


def test_calibration_coefficient_creation() -> None:
    """Check that a CalibrationCoefficientObject is correctly set up from config info."""

    c = CalibrationCoefficientItem(
        'COEF-1',
        label='Gain',
        coefficients=[100.2, 201.3],
        references=[89, 298],
        plus_tolerances=[100.2, 222.124],
        minus_tolerances=[87.23, 214],
        parent=CalibrationCoefficientSet()
    )

    assert c.name == "COEF-1"
    assert c.label.value == 'Gain'
    assert c.coefficients.value == [100.2, 201.3]
    assert c.references.value == [89, 298]
    assert c.plus_tolerances.value == [100.2, 222.124]
    assert c.minus_tolerances.value == [87.23, 214]


def test_calibration_coefficient_unequal_counts() -> None:

    c = CalibrationCoefficientItem(
        'CX',
        label='Offset',
        coefficients=[100.2, 201.3],
        references=[89, 298, 21],
        plus_tolerances=[100.2, 222.124],
        parent=CalibrationCoefficientSet()
    )

    with pytest.raises(RuntimeError,
                       match="Number of values all numeric attributes of Calibration Coefficient should be equal; "
                             "got 2 for COEFFICIENTS, 3 for REFERENCES, 2 for PLUS-TOLERANCES"):
        c.make_item_body_bytes()


@pytest.mark.parametrize("label", ("CCOEF-1", "COEFFICIENT_3", "3112-1212"))
@high_compatibility_mode_decorator
def test_ccoef_label_compatible(label: str) -> None:
    CalibrationCoefficientItem("CC-1", label=label, parent=CalibrationCoefficientSet())


@pytest.mark.parametrize("label", ("Calibration coefficient", "CALIB C0EFF-3", "C@2"))
@high_compatibility_mode_decorator
def test_ccoef_label_not_compatible(label: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        CalibrationCoefficientItem("CC-1", label=label, parent=CalibrationCoefficientSet())


@pytest.mark.parametrize("name", ("CCOEF-1", "COEFFICIENT_3", "C12"))
@high_compatibility_mode_decorator
def test_ccoef_name_compatible(name: str) -> None:
    CalibrationCoefficientItem(name, parent=CalibrationCoefficientSet())


@pytest.mark.parametrize("name", ("Coefficient", "C0EFF 3", "C.12"))
@high_compatibility_mode_decorator
def test_ccoef_name_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        CalibrationCoefficientItem(name, parent=CalibrationCoefficientSet())



from functools import lru_cache

from dliswriter.utils.internal.internal_enums import RepresentationCode


class SegmentAttributes:
    """Model metadata put in the header of each logical record segment."""

    weights = [2 ** i for i in range(8)][::-1]  # weights used to sum up the flags to represent as bytes later

    def __init__(self, is_eflr: bool = False, is_first: bool = False, is_last: bool = False):
        """Initialise SegmentAttributes.

        Args:
            is_eflr     :   True if the described logical record is an explicitly formatted one (EFLR); false otherwise.
            is_first    :   True if this is the first segment created from this logical record (no predecessors).
            is_last     :   True if this is the last segment created from this logical record (no successors).
        """

        self._value = [
            is_eflr,        #: EFLR or IFLR
            not is_first,   #: has predecessors
            not is_last,    #: has successors
            False,          #: is encrypted
            False,          #: has encryption protocol
            False,          #: has checksum
            False,          #: has trailing length
            False           #: has padding
        ]

    @property
    def has_padding(self) -> bool:
        """True if the described segment has a padding byte added; false otherwise."""

        return self._value[7]

    @has_padding.setter
    def has_padding(self, b: bool) -> None:
        """Set whether the described segment has a padding byte added."""

        self._value[7] = b

    def to_struct(self) -> bytes:
        """Transform the segment attributes to a number (by weighting and summing up flags) and that to bytes."""

        return ushort(sum(map(lambda x, y: x * y, self._value, self.weights)))


@lru_cache
def ushort(v: int) -> bytes:
    """Transform a number to bytes using USHORT format. Cache the results for future calls."""

    return RepresentationCode.USHORT.convert(v)



import pytest

from dliswriter.logical_record.eflr_types.no_format import NoFormatSet, NoFormatItem


@pytest.mark.parametrize(("name", "consumer_name", "description"), (
        ("no_format_1", "SOME TEXT NOT FORMATTED", "TESTING-NO-FORMAT"),
        ("no_fmt2", "xyz", "TESTING NO FORMAT 2")
))
def test_creation(name: str, consumer_name: str, description: str) -> None:
    """Test creating NoFormatItem."""

    w = NoFormatItem(
        name=name,
        consumer_name=consumer_name,
        description=description,
        parent=NoFormatSet()
    )

    assert w.name == name
    assert w.consumer_name.value == consumer_name
    assert w.description.value == description

    assert isinstance(w.parent, NoFormatSet)
    assert w.parent.set_name is None



import numpy as np
import pytest

from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter.utils.internal.converters import ReprCodeConverter


def _check_conversion(arr: np.ndarray, rc: RepresentationCode) -> None:
    encoded = arr.byteswap().tobytes()
    decoded_arr = np.array(rc.decode_bytes(encoded), dtype=arr.dtype)

    assert arr.size == decoded_arr.size
    assert (arr == decoded_arr).all()
    assert arr.dtype == ReprCodeConverter.determine_numpy_dtype_from_repr_code(rc)


@pytest.mark.parametrize("data", (
        np.arange(10).astype(np.float64),
        np.array([0], dtype=np.float64),
        np.random.rand(100),
        10 * np.random.rand(20) - 5
))
def test_floats(data: np.ndarray) -> None:
    _check_conversion(data.astype(np.float64), RepresentationCode.FDOUBL)
    _check_conversion(data.astype(np.float32), RepresentationCode.FSINGL)


@pytest.mark.parametrize("data", (
        np.arange(10),
        np.arange(100) - 20,
        np.array([-528]),
        np.array([20, 21, 20, 2311, -1000]),
        np.random.randint(low=-20, high=412, size=100)
))
def test_signed_ints(data: np.ndarray) -> None:
    _check_conversion(data.astype(np.int32), RepresentationCode.SLONG)
    _check_conversion(data.astype(np.int16), RepresentationCode.SNORM)
    _check_conversion(data.astype(np.int8), RepresentationCode.SSHORT)


@pytest.mark.parametrize("data", (
        np.arange(100),
        np.array([0, 20, 21, 20, 2311]),
        np.random.randint(low=0, high=1412, size=100)
))
def test_unsigned_ints(data: np.ndarray) -> None:
    _check_conversion(data.astype(np.uint32), RepresentationCode.ULONG)
    _check_conversion(data.astype(np.uint16), RepresentationCode.UNORM)
    _check_conversion(data.astype(np.uint8), RepresentationCode.USHORT)



from typing import Any, Union, Generator, Optional
from typing_extensions import Self

from dliswriter.logical_record.eflr_types.frame import FrameItem
from dliswriter.logical_record.iflr_types import FrameData
from dliswriter.utils.source_data_wrappers import SourceDataWrapper


class MultiFrameData:
    """Create a generator for FrameData objects with additional metadata and functionalities.

    Iterate over an instance of MultiFrameData to yield consecutive instances of FrameData according to the provided
    SourceDataObject (specifying numerical data, channel names, data types etc.)
    """

    def __init__(self, frame: FrameItem, data: SourceDataWrapper, chunk_size: Optional[int] = None):
        """Initialise MultiFrameData object.

        Args:
            frame       :   FrameObject instance the data refer to.
            data        :   Data (with basic metadata) to be included.
            chunk_size  :   Size (in number of rows) of chunks in which source data should be loaded when iterated over.
        """

        super().__init__()

        self._check_type(frame, FrameItem)
        self._check_type(data, SourceDataWrapper)
        self._check_type(chunk_size, int, type(None))

        frame_channel_names = tuple(c.name for c in frame.channels.value)
        data_channel_names = data.dtype.names
        if frame_channel_names != data_channel_names:
            raise ValueError(f"Channel names in data {data_channel_names} "
                             f"do not match channels defined in the frame {frame_channel_names}")

        self._data_source = data
        self._frame = frame

        self._origin_reference: Union[int, None] = self._frame.origin_reference

        self._chunk_rows = chunk_size
        self._i = 0  # keep track of current frame number during iteration
        self._data_item_generator: Union[Generator, None] = None

    @staticmethod
    def _check_type(value: Any, *expected_types: type) -> None:
        """Check that value is an instance of the expected type. If not, raise a TypeError."""

        if not isinstance(value, expected_types):
            tp = '/'.join(t.__name__ for t in expected_types)
            raise TypeError(f"Expected an instance of {tp}; got {type(value)}: {value}")

    @property
    def frame(self) -> FrameItem:
        """FrameObject the data (and FrameData objects) belong to."""

        return self._frame

    def __len__(self) -> int:
        """Number of data rows (= number of FrameData objects that can be created from the provided data)."""

        return self._data_source.n_rows

    def __iter__(self) -> Self:
        """Set up iteration over FrameData objects (to be) defined based on the data."""

        self._i = 0
        self._data_item_generator = self._data_source.make_chunked_generator(chunk_rows=self._chunk_rows)
        return self

    def __next__(self) -> FrameData:
        """Return a next FrameData object in the iteration."""

        if not self._data_item_generator:
            raise RuntimeError("Iteration has not been defined")

        if self._i >= self._data_source.n_rows:
            raise StopIteration

        self._i += 1

        return FrameData(
            frame=self._frame,
            frame_number=self._i,
            slots=next(self._data_item_generator),
            origin_reference=self._origin_reference
        )



import os
from typing import Union, Optional

from dliswriter.file import DLISFile
from dliswriter.logical_record import eflr_types
from dliswriter.utils.enums import FrameIndexType

from tests.dlis_files_for_testing.common import make_df


def _add_channels(
    df: DLISFile, lf: int, data_dict: dict, channel_kwargs: Optional[dict] = None
) -> tuple[eflr_types.ChannelItem, ...]:

    channel_kwargs = channel_kwargs or {}

    channels = []
    for key, value in data_dict.items():
        ch = df.logical_files[lf].add_channel(
            name=key, data=value, **channel_kwargs.get(key, {})
        )
        channels.append(ch)

    return tuple(channels)


def write_dlis_from_dict(
    fname: Union[str, os.PathLike[str]],
    data_dict: dict,
    channel_kwargs: Optional[dict] = None,
) -> None:
    df = make_df()

    logical_file_index = 0
    channels = _add_channels(
        df, logical_file_index, data_dict=data_dict, channel_kwargs=channel_kwargs
    )
    df.logical_files[logical_file_index].add_frame(
        "MAIN", index_type=FrameIndexType.VERTICAL_DEPTH, channels=channels
    )

    df.write(fname)



import logging
from numbers import Number
from datetime import datetime
from typing import Union, Optional, Any, overload

from .attribute import Attribute
from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import RepresentationCode as RepC
from dliswriter.utils.enums import Property
from dliswriter.utils.internal.converters import ReprCodeConverter
from dliswriter.utils.internal.types import number_type, dtime_or_number_type


logger = logging.getLogger(__name__)


class EFLRAttribute(Attribute):
    """Define an attribute whose value(s) is/are EFLRObject instances.

    Attributes like this are used when one EFLRObject instance points to another one - e.g. Zones of a Parameter
    or Channels of Frame.
    """

    _units_settable = False
    _valid_repr_codes = (RepC.OBNAME, RepC.OBJREF)
    _default_repr_code: Union[RepC, None] = RepC.OBNAME

    def __init__(self, label: str, object_class: Optional[type[EFLRSet]] = None, **kwargs: Any) -> None:
        """Initialise EFLRAttribute.

        Args:
            label               :   Attribute label.
            object_class        :   EFLR subclass corresponding to the class of the object(s) - value of the attribute.
            kwargs              :   Keyword arguments passed to Attribute.
        """

        for arg_name in ('converter', 'units'):
            if kwargs.get(arg_name, None) is not None:
                raise TypeError(f"{self.__class__.__name__} does not accept '{arg_name}' argument")

        if object_class is not None and not issubclass(object_class, EFLRSet):
            raise TypeError(f"Expected an EFLR subclass; got {object_class}")

        super().__init__(label=label, **kwargs)

        self._object_class = object_class
        self._converter = self._convert_value

    def _convert_value(self, v: EFLRItem) -> EFLRItem:
        """Implements default converter/checker for the value(s). Check that the value is an EFLRObject."""

        object_class = self._object_class.item_type if self._object_class else EFLRItem
        if not isinstance(v, object_class):
            raise TypeError(f"Expected an instance of {object_class.__name__}; got {type(v)}: {v}")
        return v


class EFLROrTextAttribute(EFLRAttribute):
    """Model an Attribute whose value might be an EFLRItem instance or a string."""

    _valid_repr_codes = (RepC.OBNAME, RepC.ASCII)
    _default_repr_code = None

    def __init__(self, label: str, **kwargs: Any):
        if kwargs.get('multivalued', False):
            raise ValueError(f"{self.__class__.__name__} cannot be multivalued")

        super().__init__(label=label, **kwargs)

    @overload
    def _convert_value(self, v: str) -> str: ...

    @overload
    def _convert_value(self, v: EFLRItem) -> EFLRItem: ...

    def _convert_value(self, v: Union[EFLRItem, str]) -> Union[EFLRItem, str]:
        """Check that the provided value is an EFLRItem or a string."""

        if isinstance(v, EFLRItem):
            return super()._convert_value(v)

        if isinstance(v, str):
            return v

        else:
            raise TypeError(f"Expected an EFLRItem or a str; got {type(v)}: {v}")

    def _guess_repr_code(self) -> Union[RepC, None]:
        """Determine representation code from the Attribute's value."""

        v = self.value

        if v is None:
            return None

        if isinstance(v, EFLRItem):
            return RepC.OBNAME

        if isinstance(v, str):
            return RepC.ASCII

        raise RuntimeError(f"Cannot determine representation code for {type(v)}: {v} in a {self.__class__.__name__}")


class DTimeAttribute(Attribute):
    """Model an attribute whose value is a datetime object."""

    dtime_formats = ["%Y/%m/%d %H:%M:%S", "%Y.%m.%d %H:%M:%S"]  #: accepted date-time formats
    _valid_repr_codes = (RepC.DTIME, RepC.FDOUBL, RepC.FSINGL)

    class DTimeFormatError(ValueError):
        """Error raised if a provided string does not match any of the allowed formats."""
        pass

    def __init__(self, *args: Any, allow_float: bool = False, **kwargs: Any) -> None:
        """Initialise a DTimeAttribute.

        Args:
            args        :   Positional arguments passed to Attribute.
            allow_float :   If True, allow for the value to be a float (e.g. a number of seconds since a specific
                            vent). Otherwise, raise an error if the user attempts to set float as the value.
            kwargs      :   Keyword arguments passed to Attribute.
        """

        super().__init__(*args, **kwargs)

        if allow_float and kwargs.get('representation_code', None) is RepC.DTIME:
            raise ValueError("Representation code cannot be specified as DTIME if float time format is allowed")

        self._allow_float = allow_float
        if not self._allow_float and not self._representation_code:
            self._representation_code = RepC.DTIME
        if not self._converter:
            self._converter = self._convert_value

    def _convert_value(self, value: Union[str, datetime, int, float]) -> dtime_or_number_type:
        """Default value converter: parse string as date time or (if so specified at init) as a float."""

        if isinstance(value, datetime):
            return value

        if isinstance(value, Number) and self._allow_float:
            return float(value)

        if not isinstance(value, str):
            raise TypeError(f"Expected a datetime object, number, or str; got {type(value)}: {value}")

        try:
            return self.parse_dtime(value)
        except self.DTimeFormatError as exc:
            if self._allow_float:
                return float(value)
            else:
                raise exc

    @classmethod
    def parse_dtime(cls, dtime_string: str) -> datetime:
        """Parse a string to a datetime object if it matches any of the allowed formats."""

        if not isinstance(dtime_string, str):
            raise TypeError(f"Expected a str, got {type(dtime_string)}")

        for dtime_format in cls.dtime_formats:
            try:
                dtime = datetime.strptime(dtime_string, dtime_format)
            except ValueError:
                pass
            else:
                break
        else:
            # loop finished without breaking - no date format fitted to the string
            raise cls.DTimeFormatError(f"Provided date time value does not conform to any of the allowed formats: "
                                       f"{', '.join(fmt for fmt in cls.dtime_formats)}")

        return dtime


class NumericAttribute(Attribute):
    """Model an attribute which can only have numerical values."""

    _valid_repr_codes = ReprCodeConverter.numeric_codes

    def __init__(self, *args: Any, int_only: bool = False, float_only: bool = False, **kwargs: Any) -> None:
        """Initialise a NumericAttribute.

        Args:
            args        :   Positional arguments passed to Attribute.
            int_only    :   If True, allow only integer values.
            float_only  :   If True, allow only float values (not integers).
            kwargs      :   Keyword arguments passed to Attribute.
        """

        self._int_only = int_only
        self._float_only = float_only

        if self._int_only and self._float_only:
            raise ValueError("'int_only' and 'float_only' cannot both be True")

        if rc := kwargs.get('representation_code', None):
            self._check_repr_code_numeric(rc)

        super().__init__(*args, **kwargs)

        if not self._converter:
            self._converter = self._convert_number
        else:
            custom_converter = self._converter

            def converter(number: number_type) -> number_type:
                number = self._convert_number(number)
                number = custom_converter(number)
                return number

            self._converter = converter

    def _check_repr_code_numeric(self, rc: Union[RepC, None]) -> None:
        """Check that the provided representation code, if not None, is of appropriate numerical type."""

        if rc is None:
            return

        if self._int_only and rc not in ReprCodeConverter.int_codes:
            raise ValueError(f"Representation code {rc.name} is not integer")

        if self._float_only and rc not in ReprCodeConverter.float_codes:
            raise ValueError(f"Representation code {rc.name} is not float")

        if rc not in ReprCodeConverter.numeric_codes:
            raise ValueError(f"Representation code {rc.name} is not numeric")

    @staticmethod
    def _int_parser(value: number_type) -> int:
        """Parse a provided value as an integer."""

        if not isinstance(value, Number):
            raise TypeError(f"Cannot convert a {type(value)} object ({value}) to integer")

        if not float(value).is_integer():
            raise ValueError(f"{value} cannot be represented as integer")

        return int(value)

    @staticmethod
    def _float_parser(value: number_type) -> float:
        """Parse a provided value as a float."""

        if not isinstance(value, Number):
            raise TypeError(f"Cannot convert a {type(value)} object ({value}) to float")

        return float(value)

    def _convert_number(self, value: number_type) -> number_type:
        """Convert a provided value according to the attribute's representation code (or as a float)."""

        if self._int_only or self.representation_code in ReprCodeConverter.int_codes:
            return self._int_parser(value)

        return self._float_parser(value)


class DimensionAttribute(NumericAttribute):
    """Model an attribute expressing dimensions (e.g. dimension or element_limit of Channel)."""

    _units_settable = False
    _valid_repr_codes = (RepC.UVARI,)
    _default_repr_code = RepC.UVARI

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise a NumericAttribute.

        Args:
            args                :   Positional arguments passed to NumericAttribute.
            kwargs              :   Keyword arguments passed to NumericAttribute.
        """

        if 'converter' in kwargs:
            raise TypeError(f"{self.__class__.__name__} does not accept 'converter' argument")

        super().__init__(*args, int_only=True, multivalued=True, **kwargs)


class StatusAttribute(Attribute):
    """Model an attribute which can only have value 1 or 0."""

    _units_settable = False
    _valid_repr_codes = (RepC.STATUS,)
    _default_repr_code = RepC.STATUS

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise a StatusAttribute.

        Args:
            args        :   Positional arguments passed to Attribute.
            kwargs      :   Keyword arguments passed to Attribute.
        """

        if kwargs.get('converter', None) is not None:
            raise ValueError(f"{self.__class__.__name__} does not accept a 'converter' argument")

        super().__init__(*args, **kwargs, converter=self.convert_status)  # type: ignore  # converter passed once

    @staticmethod
    def convert_status(val: Union[bool, int, float]) -> int:
        """Convert a provided value to 1 or 0."""

        if isinstance(val, bool):
            return int(val)

        if isinstance(val, float) and val % 1:
            raise ValueError(f"Status cannot be a fraction; got {val}")

        val = int(val)
        if val not in (0, 1):
            raise ValueError(f"Status must be a 1 or a 0; got {val}")

        return val


class TextAttribute(Attribute):
    """Model an attribute representing text in ASCII format."""

    _units_settable = False
    _valid_repr_codes = (RepC.ASCII,)
    _default_repr_code = RepC.ASCII

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise an ASCIIAttribute.

        Args:
            args        :   Positional arguments passed to Attribute.
            kwargs      :   Keyword arguments passed to Attribute.
        """

        if kwargs.get('converter', None) is not None:
            raise ValueError(f"{self.__class__.__name__} does not accept a 'converter' argument")

        super().__init__(*args, **kwargs, converter=self._check_string)  # type: ignore  # converter passed once

    @staticmethod
    def _check_string(v: str) -> str:
        if not isinstance(v, str):
            raise TypeError(f'Expected a str, got {type(v)}: {v}')
        return v


class IdentAttribute(Attribute):
    """Model an attribute represented as IDENT."""

    _units_settable = False
    _valid_repr_codes = (RepC.IDENT,)
    _default_repr_code = RepC.IDENT

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialise an IDENTAttribute.

        Args:
            args        :   Positional arguments passed to Attribute.
            kwargs      :   Keyword arguments passed to Attribute.
        """

        super().__init__(*args, **kwargs)


class PropertiesAttribute(IdentAttribute):
    """Model an attribute representing properties of DLIS objects - Channel, Computation, and Process."""

    def __init__(self, label: str) -> None:
        super().__init__(
            label,
            multivalued=True,
            converter=Property.make_converter('property indicators')
        )



from dliswriter.file.file import DLISFile, LogicalFile
from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem, AttrSetup
from dliswriter.logical_record.core.attribute import Attribute
from dliswriter.logical_record.misc.storage_unit_label import StorageUnitLabel
from dliswriter.logical_record import eflr_types
from dliswriter.utils.internal.internal_enums import RepresentationCode
from dliswriter.utils import enums
from dliswriter.utils.high_compatibility_mode import high_compatibility_mode, high_compatibility_mode_decorator
from dliswriter.utils.source_data_wrappers import SourceDataWrapper, DictDataWrapper, NumpyDataWrapper, HDF5DataWrapper


__version__ = '1.1.0'



"""From RP66 v1:

"""

from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType, RepresentationCode as RepC
from dliswriter.logical_record.core.attribute import NumericAttribute, TextAttribute


class WellReferencePointItem(EFLRItem):
    """Model an object being part of WellReferencePoint EFLR."""

    parent: "WellReferencePointSet"

    def __init__(self, name: str, parent: "WellReferencePointSet", **kwargs: Any) -> None:
        """Initialise WellReferencePointItem.

        Args:
            name        :   Name of the WellReferencePointItem.
            parent      :   Parent WellReferencePointSet of this WellReferencePointItem.
            **kwargs    :   Values of to be set as characteristics of the WellReferencePointItem Attributes.
        """

        self.permanent_datum = TextAttribute('permanent_datum')
        self.vertical_zero = TextAttribute('vertical_zero')
        self.permanent_datum_elevation = NumericAttribute('permanent_datum_elevation', representation_code=RepC.FDOUBL)
        self.above_permanent_datum = NumericAttribute('above_permanent_datum', representation_code=RepC.FDOUBL)
        self.magnetic_declination = NumericAttribute('magnetic_declination', representation_code=RepC.FDOUBL)
        self.coordinate_1_name = TextAttribute('coordinate_1_name')
        self.coordinate_1_value = NumericAttribute('coordinate_1_value', representation_code=RepC.FDOUBL)
        self.coordinate_2_name = TextAttribute('coordinate_2_name')
        self.coordinate_2_value = NumericAttribute('coordinate_2_value', representation_code=RepC.FDOUBL)
        self.coordinate_3_name = TextAttribute('coordinate_3_name')
        self.coordinate_3_value = NumericAttribute('coordinate_3_value', representation_code=RepC.FDOUBL)

        super().__init__(name, parent=parent, **kwargs)


class WellReferencePointSet(EFLRSet):
    """Model WellReferencePoint EFLR."""

    set_type = 'WELL-REFERENCE'
    logical_record_type = EFLRType.OLR
    item_type = WellReferencePointItem


WellReferencePointItem.parent_eflr_class = WellReferencePointSet



from contextlib import contextmanager
from dlisio import dlis    # type: ignore  # untyped library
from typing import Union, Generator
import os

from dliswriter.logical_record.core.eflr.eflr_item import EFLRItem


N_COLS = 128
path_type = Union[str, bytes, os.PathLike]


class FilesNotEqualError(RuntimeError):
    pass


@contextmanager
def load_dlis(fname: Union[str, bytes, os.PathLike]) -> Generator:
    """Load a DLIS file using dlisio. Yield the open file. Close the file on return to the context."""

    with dlis.load(fname) as (f, *tail):
        try:
            yield f
        finally:
            pass


def select_channel(f: dlis.file.LogicalFile, name: str) -> dlis.channel.Channel:
    """Search for a channel with given name in the dlis file and return it."""

    return f.object("CHANNEL", name)


def check_list_of_objects(objects: Union[list[EFLRItem], tuple[EFLRItem, ...]],
                          names: Union[list[str], tuple[str, ...]]) -> None:
    """Check that names of the provided objects match the expected names (in the same order)."""

    assert len(objects) == len(names)
    for i, n in enumerate(names):
        assert objects[i].name == n



import pytest

from dliswriter.logical_record.eflr_types.computation import ComputationSet, ComputationItem
from dliswriter.logical_record.eflr_types.axis import AxisItem
from dliswriter.logical_record.eflr_types.zone import ZoneItem


@pytest.mark.parametrize(("name", "properties", "zone_names", "values"), (
        ("COMPT-1", ["LOCALLY-DEFINED", "AVERAGED"], ["Zone-1", "Zone-2", "Zone-3"], [100, 200, 300]),
        ("COMPT2", ["UNDER-SAMPLED", "AVERAGED"], ["Zone-1", "Zone-3"], [1.5, 2.5]),
        ("COMPT-X", ["OVER-SAMPLED"], ["Zone-3"], [12]),
))
def test_computation_creation(name: str, properties: list[str], zone_names: list[str], values: list, axis1: AxisItem,
                              zones: dict[str, ZoneItem]) -> None:
    """Check that ComputationObject instances are correctly created."""

    comp = ComputationItem(
        name,
        properties=properties,
        values=values,
        axis=axis1,
        zones=[v for k, v in zones.items() if k in zone_names],
        parent=ComputationSet()
    )

    assert comp.name == name
    assert comp.properties.value == properties
    assert comp.axis.value[0].name == axis1.name
    assert comp.values.value == values

    for i, n in enumerate(zone_names):
        assert comp.zones.value[i].name == n

    assert isinstance(comp.parent, ComputationSet)
    assert comp.parent.set_name is None



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.eflr_types.channel import ChannelSet
from dliswriter.logical_record.eflr_types.zone import ZoneSet
from dliswriter.logical_record.core.attribute import EFLRAttribute


logger = logging.getLogger(__name__)


class SpliceItem(EFLRItem):
    """Model an object being part of Splice EFLR."""

    parent: "SpliceSet"

    def __init__(self, name: str, parent: "SpliceSet", **kwargs: Any) -> None:
        """Initialise SpliceItem.

        Args:
            name        :   Name of the SpliceItem.
            parent      :   Parent SpliceSet of this SpliceItem.
            **kwargs    :   Values of to be set as characteristics of the SpliceItem Attributes.
        """

        self.output_channel = EFLRAttribute('output_channel', object_class=ChannelSet)
        self.input_channels = EFLRAttribute('input_channels', object_class=ChannelSet, multivalued=True)
        self.zones = EFLRAttribute('zones', object_class=ZoneSet, multivalued=True)

        super().__init__(name, parent=parent, **kwargs)

    def _run_checks_and_set_defaults(self) -> None:
        """Check that the number of input channels and zones in the splice is the same."""

        if self.input_channels.value is not None and self.zones.value is not None:
            if (nc := self.input_channels.count) != (nz := self.zones.count):
                raise RuntimeError("A Splice must have the same number of input channels and zones if both are "
                                   f"defined; got {nc} channels and {nz} zones in {self}")


class SpliceSet(EFLRSet):
    """Model Splice EFLR."""

    set_type = 'SPLICE'
    logical_record_type = EFLRType.STATIC
    item_type = SpliceItem


SpliceItem.parent_eflr_class = SpliceSet



from datetime import datetime
import pytest

from dliswriter import AttrSetup, enums, high_compatibility_mode_decorator
from dliswriter.logical_record.eflr_types import (AxisItem, ChannelItem, CalibrationMeasurementItem,
                                                  CalibrationMeasurementSet)


def test_calibration_measurement_creation(channel1: ChannelItem, axis1: AxisItem) -> None:
    m = CalibrationMeasurementItem(
        "CMEASURE-1",
        **{
            'phase': enums.CalibrationMeasurementPhase.BEFORE,
            'measurement_source': channel1,
            'type': 'Plus',
            'axis': axis1,
            'measurement': AttrSetup(12.2323),
            'sample_count': {'value': 12},
            'maximum_deviation': 2.2324,
            'begin_time': '2050/03/12 12:30:00',
            'duration': AttrSetup(15, 's'),
            'reference': [11],
            'standard': [11.2],
            'plus_tolerance': [2],
            'minus_tolerance': AttrSetup(value=1),
        },
        parent=CalibrationMeasurementSet()
    )

    assert m.name == "CMEASURE-1"
    assert m.phase.value == 'BEFORE'
    assert isinstance(m.measurement_source.value, ChannelItem)
    assert m.measurement_source.value.name == "Channel 1"
    assert m.type.value == 'Plus'
    assert isinstance(m.axis.value[0], AxisItem)
    assert m.axis.value[0].name == 'Axis-1'
    assert m.measurement.value == [12.2323]
    assert m.sample_count.value == 12
    assert m.maximum_deviation.value == [2.2324]
    assert isinstance(m.begin_time.value, datetime)
    assert m.begin_time.value == datetime(2050, 3, 12, 12, 30)
    assert m.duration.value == 15
    assert m.duration.units == 's'
    assert m.reference.value == [11]
    assert m.standard.value == [11.2]
    assert m.plus_tolerance.value == [2]
    assert m.minus_tolerance.value == [1]


@pytest.mark.parametrize("t", ("PLUS", "TWO-POINT-LINEAR", "A3"))
@high_compatibility_mode_decorator
def test_type_compatible(t: str) -> None:
    CalibrationMeasurementItem("CM-1", type=t, parent=CalibrationMeasurementSet())


@pytest.mark.parametrize("t", ("Plus", "type 2", "T.1"))
@high_compatibility_mode_decorator
def test_type_not_compatible(t: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        CalibrationMeasurementItem("CM-1", type=t, parent=CalibrationMeasurementSet())


@pytest.mark.parametrize("name", ("CMEAS-1", "MEASUREMENT_3", "M11"))
@high_compatibility_mode_decorator
def test_name_compatible(name: str) -> None:
    CalibrationMeasurementItem(name, parent=CalibrationMeasurementSet())


@pytest.mark.parametrize("name", ("Meas1", "M.1", "M TYPE 2"))
@high_compatibility_mode_decorator
def test_label_not_compatible(name: str) -> None:
    with pytest.raises(ValueError, match=".*strings can contain only uppercase characters, digits, dashes, .*"):
        CalibrationMeasurementItem(name, parent=CalibrationMeasurementSet())



import pytest
from dlisio import dlis    # type: ignore  # untyped library


def test_equipment(short_dlis: dlis.file.LogicalFile) -> None:
    """Check that the number of Equipment objects in the DLIS file matches the expected one."""

    eq = short_dlis.equipments
    assert len(eq) == 3


@pytest.mark.parametrize(("idx", "name", "status", "serial_number"), (
        (0, "EQ1", 1, "9101-21391"),
        (1, "EQ2", 0, "5559101-21391"),
        (2, "EqX", 1, "12311")
))
def test_equipment_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, status: int,
                          serial_number: str) -> None:
    eq = short_dlis.equipments[idx]

    assert eq.name == name
    assert eq.status == status
    assert eq.serial_number == serial_number
    assert eq.origin == 42



from typing import Union

from dliswriter.logical_record.core.iflr import IFLR
from dliswriter.logical_record.core.logical_record.logical_record_bytes import LogicalRecordBytes
from dliswriter.utils.internal.internal_enums import IFLRType
from dliswriter.logical_record.eflr_types.no_format import NoFormatItem


class NoFormatFrameData(IFLR):
    """Model a NoFormatFrameData - an Indirectly Formatted Logical Record with data in ASCII."""

    logical_record_type = IFLRType.NOFMT

    def __init__(self, no_format_object: NoFormatItem, data: Union[str, bytes, bytearray]) -> None:
        """Initialise NoFormatFrameData."""

        super().__init__()

        self.no_format_object = no_format_object
        self.data = data

    def _make_body_bytes(self) -> bytes:
        """Create bytes representing the body of the object."""

        if isinstance(self.data, (bytes, bytearray)):
            data_encoded = self.data
        else:
            data_encoded = self.data.encode('ascii')

        bts = self.no_format_object.obname + data_encoded

        # add padding if the length of the bts is less than minimum (12)
        padding_len = 12 - len(bts)
        if padding_len:
            bts += padding_len * LogicalRecordBytes.padding

        return bts

    @property
    def n_items(self) -> int:
        """Number of data items in this object."""

        return 1



from collections.abc import Sequence
from typing import Generator, Union, Any


class SizedGenerator(Sequence):
    """Wrap a generator with a pre-defined number of elements."""

    def __init__(self, generator: Generator, size: int) -> None:
        self._generator = generator
        self._size = size

    def __len__(self) -> int:
        return self._size

    def __iter__(self) -> Generator:
        yield from self._generator

    def __getitem__(self, index: Union[int, slice]) -> Any:
        raise NotImplementedError("SizedGenerator does not support item access")



from contextlib import contextmanager
from typing import Generator, Callable, Any
import functools

from dliswriter.configuration import global_config


@contextmanager
def high_compatibility_mode() -> Generator:
    """Context manager. Turn on DLIS Writer's high-compatibility mode for the scope of the context."""

    arch = global_config.high_compat_mode
    global_config.high_compat_mode = True

    try:
        yield
    finally:
        global_config.high_compat_mode = arch


def high_compatibility_mode_decorator(func: Callable) -> Callable:
    """Decorator. Turn on DLIS Writer's high-compatibility mode for the scope of the decorated function."""

    @functools.wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> None:
        with high_compatibility_mode():
            func(*args, **kwargs)

    return wrapper



import logging
from typing import Optional

from dliswriter.utils.internal.struct_writer import write_struct_ascii
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.logical_record import LogicalRecord
from dliswriter.logical_record.core.eflr.eflr_item import EFLRItem


logger = logging.getLogger(__name__)


class EFLRSet(LogicalRecord):
    """Model an Explicitly Formatted Logical Record."""

    set_type: str = NotImplemented                  #: set type of each particular EFLR (e.g. Channel); see the standard
    logical_record_type: EFLRType = NotImplemented  #: int-enum denoting type of the EFLR
    is_eflr: bool = True                            #: indication that this is an explicitly formatted LR
    item_type: type[EFLRItem] = EFLRItem            #: EFLRItem subclass which can be held within this EFLRTable type

    def __init__(self, set_name: Optional[str] = None):
        """Initialise an EFLRTable.

        Args:
            set_name    :   Name of the set this EFLR belongs to. Multiple EFLRSet instances of the same type
                            (subclass) can be included in the same file if their set names differ.
        """

        super().__init__()

        self.set_name = set_name
        self._set_type_struct = write_struct_ascii(self.set_type)  # used in the header
        self._eflr_item_list: list[EFLRItem] = []  # instances of EFLRItem registered with this EFLRSet instance

    def __str__(self) -> str:
        """Represent the EFLRSet instance as str."""

        return f"{self.__class__.__name__} {repr(self.set_name)}"

    def _make_set_component_bytes(self) -> bytes:
        """Create bytes describing the set of this EFLR, using set type (class attr) and name (specified at init)."""

        if self.set_name:
            _bytes = b'\xf8' + self._set_type_struct + write_struct_ascii(self.set_name)
        else:
            _bytes = b'\xf0' + self._set_type_struct

        return _bytes

    def _make_template_bytes(self) -> bytes:
        """Create bytes describing the attribute template of this EFLR.

        Note: if no EFLRItems are registered, this will return an empty bytes object."""

        _bytes = b''
        if self._eflr_item_list:
            child0 = self._eflr_item_list[0]
            for attr in child0.attributes.values():
                _bytes += attr.get_as_bytes(for_template=True)

        return _bytes

    def _make_body_bytes(self) -> bytes:
        """Create bytes describing the body of this EFLRSet - the values of attributes of the registered EFLRItems.

        If no EFLRItems are registered, this will return an empty bytes object.
        """

        eflr_items = self.get_all_eflr_items()
        if not eflr_items:
            return b''

        bts = self._make_set_component_bytes() + self._make_template_bytes()
        for ei in eflr_items:
            bts += ei.make_item_body_bytes()

        return bts

    def register_item(self, child: EFLRItem) -> None:
        """Register a child EFLRItem with this EFLRSet."""

        if not isinstance(child, self.item_type):
            raise TypeError(f"Expected an instance of {self.item_type}; got {type(child)}: {child}")

        self._eflr_item_list.append(child)

    def get_all_eflr_items(self) -> list[EFLRItem]:
        """Return a list of all EFLRItem instances registered with this EFLRSet instance."""

        return self._eflr_item_list[:]  # copy

    @property
    def n_items(self) -> int:
        """Number of EFLRItem instances registered with this EFLRSet instance."""

        return len(self._eflr_item_list)



import logging
from typing import Any

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.logical_record.eflr_types.frame import FrameSet
from dliswriter.logical_record.eflr_types.channel import ChannelSet
from dliswriter.logical_record.eflr_types.well_reference_point import WellReferencePointSet
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.logical_record.core.attribute import EFLRAttribute, NumericAttribute


logger = logging.getLogger(__name__)


class PathItem(EFLRItem):
    """Model an object being part of Path EFLR."""

    parent: "PathSet"

    def __init__(self, name: str, parent: "PathSet", **kwargs: Any) -> None:
        """Initialise PathItem.

        Args:
            name        :   Name of the PathItem.
            parent      :   Parent PathSet of this PathItem.
            **kwargs    :   Values of to be set as characteristics of the PathItem Attributes.
        """

        self.frame_type = EFLRAttribute('frame_type', object_class=FrameSet)
        self.well_reference_point = EFLRAttribute('well_reference_point', object_class=WellReferencePointSet)
        self.value = EFLRAttribute('value', object_class=ChannelSet, multivalued=True)
        self.borehole_depth = NumericAttribute('borehole_depth')
        self.vertical_depth = NumericAttribute('vertical_depth')
        self.radial_drift = NumericAttribute('radial_drift')
        self.angular_drift = NumericAttribute('angular_drift')
        self.time = NumericAttribute('time')
        self.depth_offset = NumericAttribute('depth_offset')
        self.measure_point_offset = NumericAttribute('measure_point_offset')
        self.tool_zero_offset = NumericAttribute('tool_zero_offset')

        super().__init__(name, parent=parent, **kwargs)


class PathSet(EFLRSet):
    """Model Path EFLR."""

    set_type = 'PATH'
    logical_record_type = EFLRType.FRAME
    item_type = PathItem


PathItem.parent_eflr_class = PathSet



import h5py    # type: ignore  # untyped library
import pytest
from pathlib import Path
import numpy as np

from tests.common import N_COLS, load_dlis, select_channel
from tests.dlis_files_for_testing import write_time_based_dlis, write_depth_based_dlis, write_dlis_from_dict


def test_dlis_depth_based(short_reference_data: h5py.File, short_reference_data_path: Path, new_dlis_path: Path)\
        -> None:
    """Create a depth-based DLIS file and check its basic parameters."""

    write_depth_based_dlis(new_dlis_path, data=short_reference_data_path)

    with load_dlis(new_dlis_path) as f:
        chan = f.channels[0]
        assert chan.name == 'depth'
        assert chan.units == 'm'
        assert chan.reprc == 7
        assert chan.origin == 1

        frame = f.frames[0]
        assert frame.index_type == 'DEPTH'
        assert frame.origin == 1

        index = short_reference_data['/contents/depth'][:]
        assert frame.index_min == index.min()
        assert frame.index_max == index.max()


def test_dlis_time_based(short_reference_data: h5py.File, short_reference_data_path: Path, new_dlis_path: Path) -> None:
    """Create a time-based DLIS file and check its basic parameters."""

    write_time_based_dlis(new_dlis_path, data=short_reference_data_path)

    with load_dlis(new_dlis_path) as f:
        chan = f.channels[0]
        assert chan.name == 'posix time'
        assert chan.units == 's'
        assert chan.reprc == 7
        assert chan.origin == 1

        frame = f.frames[0]
        assert frame.index_type == 'TIME'
        assert frame.origin == 1

        index = short_reference_data['/contents/time'][:]
        assert frame.index_min == index.min()
        assert frame.index_max == index.max()


def test_repr_code(short_reference_data_path: Path, new_dlis_path: Path) -> None:
    """Test that representation codes for the channels in DLIS are stored correctly."""

    write_time_based_dlis(new_dlis_path, data=short_reference_data_path)

    with load_dlis(new_dlis_path) as f:
        assert select_channel(f, 'posix time').reprc == 7
        assert select_channel(f, 'surface rpm').reprc == 7
        assert select_channel(f, 'amplitude').reprc == 2
        assert select_channel(f, 'radius').reprc == 2
        assert select_channel(f, 'radius_pooh').reprc == 2


def test_channel_dimensions(short_reference_data_path: Path, new_dlis_path: Path) -> None:
    """Test that dimensions for the channels in DLIS are stored correctly."""

    write_time_based_dlis(new_dlis_path, data=short_reference_data_path)

    with load_dlis(new_dlis_path) as f:
        def check(name: str, shape: list) -> None:
            ch = select_channel(f, name)
            assert ch.dimension == shape
            assert ch.element_limit == shape

        check('posix time', [1])
        check('surface rpm', [1])
        check('amplitude', [128])
        check('radius', [128])
        check('radius_pooh', [128])


@pytest.mark.parametrize('n_points', (10, 100, 128, 987))
def test_channel_curves(reference_data_path: Path, reference_data: h5py.File, new_dlis_path: Path,
                        n_points: int) -> None:
    """Create a DLIS file with varying number of points. Check that the data for each channel are correct."""

    write_time_based_dlis(new_dlis_path, data=reference_data_path, to_idx=n_points)

    with load_dlis(new_dlis_path) as f:
        for name in ('posix time', 'surface rpm'):
            curve = select_channel(f, name).curves()
            assert curve.shape == (n_points,)

        for name in ('amplitude', 'radius', 'radius_pooh'):
            curve = select_channel(f, name).curves()
            assert curve.shape == (n_points, N_COLS)

        def check_contents(channel_name: str, data_name: str) -> None:
            curve = select_channel(f, channel_name).curves()
            data = reference_data[data_name][:n_points]
            assert pytest.approx(curve) == data

        check_contents('posix time', '/contents/time')
        check_contents('surface rpm', '/contents/rpm')
        check_contents('amplitude', '/contents/image0')
        check_contents('radius', '/contents/image1')
        check_contents('radius_pooh', '/contents/image2')


@pytest.mark.parametrize('data_arr', (
    np.random.rand(100).astype(np.float64),
    np.random.rand(30, 10).astype(np.float32),
    np.random.randint(0, 2**32, size=(2, 30), dtype=np.uint32),
    np.random.randint(0, 2**16, size=(100, 5), dtype=np.uint16),
    np.random.randint(0, 2**8, size=15, dtype=np.uint8),
    np.random.randint(-2**16, 2**16, size=280, dtype=np.int32),
    np.random.randint(-2**15, 2**15, size=33, dtype=np.int16),
    np.random.randint(-2**7, 2**7, size=(12, 13), dtype=np.int8)
))
def test_all_numpy_dtypes(new_dlis_path: Path, data_arr: np.ndarray) -> None:

    data_arr = np.atleast_2d(data_arr)
    data_dict = {
        'index': np.arange(data_arr.shape[0]).astype(np.float64),
        'data': data_arr
    }

    write_dlis_from_dict(new_dlis_path, data_dict=data_dict)

    with load_dlis(new_dlis_path) as f:
        ch = select_channel(f, 'data')
        assert ch.dimension == [data_arr.shape[-1]]
        assert ch.curves().dtype == data_arr.dtype
        assert (ch.curves() == data_arr).all()


@pytest.mark.parametrize(('data_arr', 'cast_dtype', 'rc'), (
    (np.random.randint(-10, 30, (10, 3)), np.float64, 7),
    (np.random.rand(30, 10) + 20, np.float32, 2),
    (np.random.randint(0, 2**16, size=(2, 30), dtype=np.uint16), np.uint32, 17),
    (1024 * np.random.rand(100), np.uint16, 16),
    (30 * np.random.rand(128, 10), np.uint8, 15),
    (np.random.randint(-2**16, 2**16, size=100), np.int32, 14),
    (100 * np.random.rand(20) - 30, np.int16, 13),
    (50 * np.random.rand(12, 13) - 40, np.int8, 12)
))
def test_numpy_dtypes_with_casting(new_dlis_path: Path, data_arr: np.ndarray, cast_dtype: type[np.generic],
                                   rc: int) -> None:

    data_arr = np.atleast_2d(data_arr)
    data_dict = {
        'index': np.arange(data_arr.shape[0]).astype(np.float64),
        'data': data_arr
    }

    write_dlis_from_dict(new_dlis_path, data_dict=data_dict, channel_kwargs={'data': {'cast_dtype': cast_dtype}})

    with load_dlis(new_dlis_path) as f:
        ch = select_channel(f, 'data')
        assert ch.dimension == [data_arr.shape[-1]]
        assert ch.reprc == rc
        assert ch.curves().dtype == cast_dtype
        assert (ch.curves() == data_arr.astype(cast_dtype)).all()



from typing import Any
import logging

from dliswriter.logical_record.core.eflr import EFLRSet, EFLRItem
from dliswriter.utils.internal.internal_enums import EFLRType
from dliswriter.utils.enums import EquipmentType, EquipmentLocation
from dliswriter.logical_record.core.attribute import NumericAttribute, StatusAttribute, TextAttribute, IdentAttribute
from dliswriter.utils.internal.value_checkers import validate_string


logger = logging.getLogger(__name__)


class EquipmentItem(EFLRItem):
    """Model an object being part of Equipment EFLR."""

    parent: "EquipmentSet"

    def __init__(self, name: str, parent: "EquipmentSet", **kwargs: Any) -> None:
        """Initialise EquipmentItem.

        Args:
            name        :   Name of the EquipmentItem.
            parent      :   Parent EquipmentSet of this EquipmentItem.
            **kwargs    :   Values of to be set as characteristics of the EquipmentItem Attributes.
        """

        self.trademark_name = TextAttribute('trademark_name')
        self.status = StatusAttribute('status')
        self._type = IdentAttribute(
            '_type', converter=EquipmentType.make_converter("equipment types", soft=True))
        self.serial_number = IdentAttribute('serial_number', converter=validate_string)
        self.location = IdentAttribute(
            'location', converter=EquipmentLocation.make_converter("equipment locations", soft=True))
        self.height = NumericAttribute('height')
        self.length = NumericAttribute('length')
        self.minimum_diameter = NumericAttribute('minimum_diameter')
        self.maximum_diameter = NumericAttribute('maximum_diameter')
        self.volume = NumericAttribute('volume')
        self.weight = NumericAttribute('weight')
        self.hole_size = NumericAttribute('hole_size')
        self.pressure = NumericAttribute('pressure')
        self.temperature = NumericAttribute('temperature')
        self.vertical_depth = NumericAttribute('vertical_depth')
        self.radial_drift = NumericAttribute('radial_drift')
        self.angular_drift = NumericAttribute('angular_drift')

        super().__init__(name, parent=parent, **kwargs)


class EquipmentSet(EFLRSet):
    """Model Equipment EFLR."""

    set_type = 'EQUIPMENT'
    logical_record_type = EFLRType.STATIC
    item_type = EquipmentItem


EquipmentItem.parent_eflr_class = EquipmentSet



import os
from pathlib import Path
import logging
import coloredlogs  # type: ignore  # untyped

from dliswriter.utils.internal.types import file_name_type


logger = logging.getLogger(__name__)


def _check_write_access(p: file_name_type) -> None:
    """Check if the provided path supports write action. Raise a RuntimeError otherwise."""

    if not os.access(p, os.W_OK):
        raise RuntimeError(f"Write permissions missing for directory: {p}")


def prepare_directory(output_file_name: file_name_type, overwrite: bool = False) -> None:
    """Prepare directory for the output file.

    Create up to 1 top level on the path. Make sure the directory allows writing.
    Check if a file of the given name already exists.

    Args:
        output_file_name    :   Name of the file to be created.
        overwrite           :   Used if the file already exists. If True, include a warning in the logs and overwrite
                                the file. If False, raise a RuntimeError.
    """

    output_file_name = Path(output_file_name).resolve()
    save_dir = output_file_name.parent
    parent_dir = save_dir.parent

    if not parent_dir.exists():
        raise RuntimeError(f"Directory {parent_dir} does not exist")

    _check_write_access(parent_dir)

    os.makedirs(save_dir, exist_ok=True)
    _check_write_access(save_dir)

    if os.path.exists(output_file_name):
        if overwrite:
            logger.warning(f"Output file at {output_file_name} will be overwritten")
        else:
            raise RuntimeError(f"Cannot overwrite existing file at {output_file_name}")


def install_colored_logger(log: logging.Logger, level: int = logging.DEBUG) -> None:
    """Set up a colored logging output.

    Args:
        log     :   The logger to install as a colored logger.
        level   :   The minimum level of messages for them to be displayed in the console logs.

    Both arguments are passed to 'coloredlogs.install'.
    """

    coloredlogs.install(
        logger=log,
        fmt='%(name)s [%(levelname)s] %(asctime)s: %(message)s',
        level=level,
        level_styles={
            'debug': {'color': 35},
            'info': {'color': 74},
            'warning': {'color': 190},
            'error': {'color': 124}
        },
        field_styles={
            'asctime': {'color': 102},
            'name': {'color': 102},
            'levelname': {'color': 231}
        }
    )



import pytest
from datetime import datetime, timezone
from dlisio import dlis    # type: ignore  # untyped library


def test_message_params(short_dlis: dlis.file.LogicalFile) -> None:
    """Check attribute of Message object in the DLIS file."""

    m = short_dlis.messages[0]

    assert m.name == "MESSAGE-1"
    assert m.message_type == 'Command'
    assert m.time.replace(tzinfo=timezone.utc) == datetime(2050, 3, 4, 11, 23, 11).astimezone(timezone.utc)
    assert m.borehole_drift == 123.34
    assert m.vertical_depth == 234.45
    assert m.radial_drift == 345.56
    assert m.angular_drift == 456.67
    assert m.text == ["Test message 11111"]
    assert m.origin == 42


@pytest.mark.parametrize(("idx", "name", "text"), (
        (0, "COMMENT-1", ["SOME COMMENT HERE"]),
        (1, "cmt2", ["some other comment here", "and another comment"])
))
def test_comment_params(short_dlis: dlis.file.LogicalFile, idx: int, name: str, text: str) -> None:
    """Test attributes of Comment objects in the DLIS file."""

    c = short_dlis.comments[idx]

    assert c.name == name
    assert c.text == text
    assert c.origin == 42



from dliswriter.logical_record.eflr_types.axis import AxisSet, AxisItem
from dliswriter.logical_record.eflr_types.calibration_measurement import (CalibrationMeasurementItem,
                                                                          CalibrationMeasurementSet)
from dliswriter.logical_record.eflr_types.calibration_coefficient import (CalibrationCoefficientSet,
                                                                          CalibrationCoefficientItem)
from dliswriter.logical_record.eflr_types.calibration import CalibrationSet, CalibrationItem
from dliswriter.logical_record.eflr_types.channel import ChannelSet, ChannelItem
from dliswriter.logical_record.eflr_types.comment import CommentSet, CommentItem
from dliswriter.logical_record.eflr_types.computation import ComputationSet, ComputationItem
from dliswriter.logical_record.eflr_types.equipment import EquipmentSet, EquipmentItem
from dliswriter.logical_record.eflr_types.frame import FrameSet, FrameItem
from dliswriter.logical_record.eflr_types.group import GroupSet, GroupItem
from dliswriter.logical_record.eflr_types.long_name import LongNameSet, LongNameItem
from dliswriter.logical_record.eflr_types.message import MessageSet, MessageItem
from dliswriter.logical_record.eflr_types.no_format import NoFormatSet, NoFormatItem
from dliswriter.logical_record.eflr_types.origin import OriginSet, OriginItem
from dliswriter.logical_record.eflr_types.parameter import ParameterSet, ParameterItem
from dliswriter.logical_record.eflr_types.path import PathSet, PathItem
from dliswriter.logical_record.eflr_types.process import ProcessSet, ProcessItem
from dliswriter.logical_record.eflr_types.splice import SpliceSet, SpliceItem
from dliswriter.logical_record.eflr_types.tool import ToolSet, ToolItem
from dliswriter.logical_record.eflr_types.well_reference_point import WellReferencePointSet, WellReferencePointItem
from dliswriter.logical_record.eflr_types.zone import ZoneSet, ZoneItem

from dliswriter.logical_record.eflr_types.file_header import FileHeaderSet, FileHeaderItem


eflr_sets = (
    AxisSet,
    CalibrationSet,
    CalibrationMeasurementSet,
    CalibrationCoefficientSet,
    ChannelSet,
    ComputationSet,
    EquipmentSet,
    FrameSet,
    GroupSet,
    LongNameSet,
    MessageSet,
    CommentSet,
    NoFormatSet,
    OriginSet,
    ParameterSet,
    PathSet,
    ProcessSet,
    SpliceSet,
    ToolSet,
    WellReferencePointSet,
    ZoneSet,
    FileHeaderSet
)


eflr_items = (
    AxisItem,
    CalibrationItem,
    CalibrationMeasurementItem,
    CalibrationCoefficientItem,
    ChannelItem,
    ComputationItem,
    EquipmentItem,
    FrameItem,
    GroupItem,
    LongNameItem,
    MessageItem,
    CommentItem,
    NoFormatItem,
    OriginItem,
    ParameterItem,
    PathItem,
    ProcessItem,
    SpliceItem,
    ToolItem,
    WellReferencePointItem,
    ZoneItem,
    FileHeaderItem
)
